import numpy as np

from dataclasses import dataclass
from scipy.stats import multivariate_normal, norm

from ....settings.settings import Settings
from .....io.database_hr import DatabaseHR


class ModelUncertainty:
    """
    Model uncertainties class. Containing all model uncertainties for each closing situation.

    Attributes
    ----------
    model_uncertainties : dict
        A dictionary with
    """

    # Init attributes
    model_uncertainties = {}
    correlations = {}

    def __init__(self, settings: Settings):
        """
        Read the model uncertainties from the database and add them to this class

        Parameters
        ----------
        settings : Settings
            The Settings object
        """
        # Save the discretisation step size
        self.step_size = {
            "h": settings.model_uncertainty_water_level_steps,
            "hs": settings.model_uncertainty_wave_height_steps,
            "tspec": settings.model_uncertainty_wave_period_steps,
        }

        # Obtain the model uncertainties and correlation between model uncertainties
        with DatabaseHR(settings.database_path) as database:
            mu = database.get_model_uncertainties(settings)
            cu = database.get_correlation_uncertainties(settings)

        # Iterate over the model uncertainties and add them to the class
        for comb, uncertainty in mu.groupby(["k", "rvid"]):
            self.model_uncertainties[comb] = DistributionUncertainty(
                uncertainty.to_numpy()[0]
            )

        # Iterate over the correlation between model uncertainties and add them to the class
        for comb, correlation in cu.groupby(["k", "rvid", "rvid2"]):
            self.correlations[comb] = CorrelationUncertainty(correlation.to_numpy()[0])

    def iterate_model_uncertainty_wave_conditions(
        self, closing_situation: int = 1, wave_period: str = "tspec"
    ):
        """
        Iterate over all model uncertainty combinations for the significant wave height and wave period

        Parameters
        ----------
        closing_situation : int, optional
            The closing situation id (default : 1)
        wave_period : str, optional
            Whether to iterate over peak period (tp) or spectral wave period (tspec) (default : 'tspec')

        Returns
        -------
        iterator
            (hs, t, probability of (hs, t))
        """
        # Check if tp or tspec
        wave_period = wave_period.lower()
        if wave_period not in ["tp", "tspec"]:
            raise KeyError(f"[ERROR] Wave period '{wave_period}' unknown.")

        # Distributions
        mu_hs = self.get_model_uncertainty("hs", closing_situation)
        mu_t = self.get_model_uncertainty(wave_period, closing_situation)

        # Significant wave height
        if mu_hs is not None:
            hs, hsedges = mu_hs.discretise(self.step_size["hs"])
            tmp = norm.cdf(hsedges)
            p_hs = tmp[1:] - tmp[:-1]
        else:
            hs, p_hs = [1.0], np.array([1.0])

        # Wave period
        if mu_t is not None:
            t, tedges = mu_t.discretise(self.step_size["tspec"])
            tmp = norm.cdf(tedges)
            p_t = tmp[1:] - tmp[:-1]
        else:
            t, p_t = [1.0], np.array([1.0])

        # Multiply the probabilities (assuming independence)
        combined_probability = p_hs[:, None] * p_t[None, :]

        # Obtain the correlation between wave height and period (optional, otherwise None)
        corr_hs_t = self.get_correlation("hs", wave_period)

        # If rho is defined, apply correlation
        if corr_hs_t is not None:
            # Correlation can only be applied when both the wave height and period are defined
            if (mu_hs is not None) and (mu_t is not None):
                # Apply correlation
                exc_probs = np.zeros((len(hsedges), len(tedges)))
                for i, x in enumerate(hsedges):
                    for j, y in enumerate(tedges):
                        exc_probs[i, j] = multivariate_normal.cdf(
                            [x, y],
                            mean=(0, 0),
                            cov=[[1, corr_hs_t.rho], [corr_hs_t.rho, 1]],
                        )

                # Take the difference in both directions
                combined_probability = exc_probs[1:] - exc_probs[:-1]
                combined_probability = (
                    combined_probability[:, 1:] - combined_probability[:, :-1]
                )

            # Otherwise give a warning
            else:
                print(
                    "[WARNING] Correlation between wave height and period defined. However can not be applied because no model uncertainty is defined for either the wave height, period or both."
                )

        # Check
        assert abs(combined_probability.sum() - 1) < 1e-6

        # Iterator
        for i, fh in enumerate(hs):
            for j, ft in enumerate(t):
                yield fh, ft, combined_probability[i, j]

    def get_model_uncertainty(self, result_variable: str, closing_situation: int = 1):
        """
        Return the model uncertainty object for a result variable and closing situation id

        Parameters
        ----------
        result_variable : str
            Result variable (e.g. h, hs)
        closing_situation : int
            Closing situation ID (default: 1)

        Returns
        -------
        DistributionUncertainty or None
            The DistributionUncertainty object if it exists, otherwise None
        """
        # To lower
        rv = result_variable.lower()

        # Try to return the model uncertainty, otherwise return None
        return self.model_uncertainties.get((closing_situation, rv))

    def get_correlation(
        self, result_variable1: str, result_variable2: str, closing_situation: int = 1
    ):
        """
        Return the correlation object between two result variables given a closing situation id

        Parameters
        ----------
        result_variable1 : str
            Result variable (e.g. h, hs)
        result_variable2 : str
            Result variable (e.g. h, hs)
        closing_situation : int
            Closing situation ID (default: 1)

        Returns
        -------
        Correlation or None
            The Correlation object if it exists, otherwise None
        """
        # To lower
        rv1 = result_variable1.lower()
        rv2 = result_variable2.lower()

        # Check both orders [ccid, rvid1, rvid2]
        return self.correlations.get(
            (closing_situation, rv1, rv2),
            self.correlations.get((closing_situation, rv2, rv1)),
        )

    def process_model_uncertainty(
        self,
        closing_situation: int,
        result_variable: str,
        levels: np.ndarray,
        exceedance_probability: np.ndarray,
        haxis: int,
    ):
        """
        Verwerk modelonzekerheid in gegeven stochast. Afhankelijk van de stochast wordt de onzekerheid
        opgeteld (additief) of vermenigvuldigd (multiplicatief).

        Args:
            stochast (str): Stochastnaam. h, hs, tp of tspec
            niveaus (np.ndarray): Niveaus waarin de onzekerheden worden ingedeeld.
            ovkansen (np.ndarray): Overschrijdingskansen van de stochast, met eventueel meerdere dimensies
            die meegeïntegreerd
            haxis (int): as waarop de te integreren stochast zit
            sluitsituatie (str): Keringsituatie, in sommige gevallen is de onzekerheid hiervan afhankelijk

        Returns:
            np.ndarray: Overschrijdingskansen met geïntegreerde onzekerheid
        """
        # Obtain distribution
        dis = self.get_model_uncertainty(result_variable, closing_situation)

        # Is the model uncertainty defined?
        if dis is None:
            return exceedance_probability

        # Additive (h) or Multiplicative (hs, tspec, tp)
        if result_variable in ["h"]:
            klassekansen = self.bepaal_klassekansen_additief(levels, dis.mu, dis.sigma)

        elif result_variable in ["hs", "tp", "tspec"]:
            klassekansen = self.bepaal_klassekansen_multiplicatief(
                levels, dis.mu, dis.sigma
            )

        else:
            raise KeyError(result_variable)

        # Calculate the exceedance probability
        exceedance_probability = np.tensordot(
            klassekansen, exceedance_probability, axes=([0], [haxis])
        )

        return exceedance_probability

    def bepaal_klassekansen_additief(self, niveaus, mu, sigma):
        # Bepaal klassegrenzen en klassekansen
        hgrens = np.concatenate(
            [[-np.inf], (niveaus[1:] + niveaus[:-1]) / 2.0, [np.inf]]
        )
        klassekansen = []

        # Bereken per niveau (waterstand) de kans dat de waterstand door onzekerheid in een andere klasse valt
        for niveau in niveaus:
            grenskansen = norm.cdf(x=hgrens - mu, loc=niveau, scale=sigma)
            klassekansen.append(grenskansen[1:] - grenskansen[:-1])
        klassekansen = np.array(klassekansen)

        return klassekansen

    def bepaal_klassekansen_multiplicatief(self, niveaus, mu, sigma):
        # Bepaal klassegrenzen en klassekansen
        hgrens = np.concatenate(
            [[-np.inf], (niveaus[1:] + niveaus[:-1]) / 2.0, [np.inf]]
        )
        klassekansen = []

        # Bereken per niveau (waterstand) de kans dat de waterstand door onzekerheid in een andere klasse valt
        for niveau in niveaus:
            grenskansen = norm.cdf(hgrens / niveau, loc=mu, scale=sigma)
            klassekansen.append(grenskansen[1:] - grenskansen[:-1])
        klassekansen = np.array(klassekansen)

        return klassekansen


class DistributionUncertainty:
    """
    Model uncertainty class for a closing situation.
    """

    def __init__(self, uncertainty: list):
        """
        Initialise the model uncertainty (Normal Distribution)
        """
        # Save information
        self.k, self.rvid, self.mu, self.sigma = uncertainty
        self.k = int(self.k)
        self.mu = float(self.mu)
        self.sigma = float(self.sigma)

    def discretise(self, nsteps: int):
        """
        Discretise the model uncertainty

        Parameters
        ----------
        nsteps : int
            Number of steps

        Returns
        -------
        probabilities : list
            List of the probabilities for each bin
        edges : list
            List with the edges of each bin
        """
        # If nsteps is 1, there is no need to discretise
        if nsteps == 1:
            return self.mu, [self.mu - 100 * self.sigma, self.mu + 100 * self.sigma]

        # Determine the residual probabilities
        keuzerestkans = 0.05
        restkans = keuzerestkans / (nsteps**1.5)
        afstand = -norm.ppf(q=0.5 * restkans, loc=0.0, scale=1.0) * self.sigma
        ondergrens = self.mu - afstand
        bovengrens = self.mu + afstand

        # Calculate the probability at the center of the bin
        probabilities = (
            ondergrens + np.arange(0.5, nsteps, 1) * (bovengrens - ondergrens) / nsteps
        )

        # Determine the edges of the bins
        edges = (
            np.concatenate(
                [
                    [self.mu - 100 * self.sigma],
                    (probabilities[1:] + probabilities[:-1]) / 2,
                    [self.mu + 100 * self.sigma],
                ]
            )
            - self.mu
        ) / self.sigma

        # Return probabilities and edges
        return probabilities, edges


@dataclass
class CorrelationUncertainty:
    """
    Class om de correlaties tussen uitvoervariabelen op te slaan.
    Bijv. de correlatie tussen Hs en Tspec
    """

    def __init__(self, correlation: list):
        """
        Initialise the correlation between two result variables
        """
        # Save information
        self.k, self.rvid, self.rvid2, self.rho = correlation
        self.k = int(self.k)
        self.rho = float(self.rho)

        # Only allow correlation between Hs and Tp and Hs and Tspec
        if [self.rvid, self.rvid2] not in [
            ["hs", "tp"],
            ["tp", "hs"],
            ["hs", "tspec"],
            ["tspec", "hs"],
        ]:
            raise ValueError(
                f"Not Implemented: Correlation between ({self.rvid}) and ({self.rvid2})"
            )



from ..loading import Loading
from ..loading_model.loading_model import LoadingModel
from ....settings.settings import Settings
from .....io.database_hr import DatabaseHR


class LoadingLake(Loading):
    """
    Loading class for the Lakes
    Water systems: IJssel Lake, Marker Lake, Veluwe Lakes and Grevelingen
    """

    def __init__(self, settings: Settings):
        """
        Init the Loading object for the Upper Rivers

        Parameters
        ----------
        settings : Settings
            The Settings object
        """
        # Inherit the from parent
        super().__init__(settings)

        # Read and process the loading
        self.read_loading()

    def read_loading(self) -> None:
        """
        Read the HR result table and create LoadingModels
        """
        # Read table
        with DatabaseHR(self.settings.database_path) as database:
            table = database.get_result_table(self.settings)
            ivids = database.get_input_variables()
            rvids = database.get_result_variables()

        # Check if there are wave conditions present or whether they should be derived with Bretschneider
        if "hs" not in table.columns:
            raise NotImplementedError("[ERROR] Bretschneider")

        # Init LoadingModels for each combination of wind direction (r) and closing situation (k)
        for comb, deeltabel in table.groupby(["r", "k"]):
            direction, closing_situation = comb

            # Create a LoadingModel
            model = LoadingModel(direction, closing_situation, ivids, rvids)
            model.initialise(deeltabel.copy())

            # Add model to the models dictionary
            self.model[comb] = model

        # Extend and repair loadingmodels
        self._extend_loadingmodels()
        self.repair_loadingmodels(rvids)



import numpy as np
import pandas as pd

from typing import Dict, Union

from ....profile.profile import Profile
from .....common.interpolate import InterpStruct


class LoadingModel:
    """
    A LoadingModel is a model for one combination of wind direction and closing situation

    The LoadingModel allows to process the data (e.g. extend, refine or repair)
    """

    def __init__(
        self,
        direction: float,
        closing_situation: int,
        input_variables: list,
        result_variables: list,
    ):
        """
        Init the LoadingModel.

        Parameters
        ----------
        direction : float
            Wind direction
        closing_situation: int
            Closing situation id
        input_variables : list
            The input variable symbols (e.q. [u, q])
        result_variables : list
            The result variable symbols (e.q. [h, hs, tp, tspec, dir])
        """
        # Save the arguments into the object
        self.direction = direction
        self.closing_situation = closing_situation
        self.input_variables = input_variables
        self.result_variables = result_variables

    def initialise(self, table: pd.DataFrame) -> None:
        """
        Create a LoadingModel from a pandas DataFrame.

        Parameters
        ----------
        table : pd.DataFrame
            DataFrame with all input and output variables.
        """
        # Rename columns
        if any(ivar not in table.columns for ivar in self.input_variables):
            raise KeyError(
                f"Not all input variables are present. Expected a column for each of: {', '.join(self.input_variables)}, got: {', '.join(table.columns.tolist())}."
            )

        # Add the discretisation of each input variable to the object
        for key in self.input_variables:
            setattr(self, key, np.sort(table[key].unique()))

        # Init an empty grid for each result variable
        shape = tuple([len(getattr(self, key)) for key in self.input_variables])
        for var in self.result_variables:
            setattr(self, var, np.full(shape, np.nan))

        # Determine per input variabele where to put it into the results array
        idxs = []
        for key in self.input_variables:
            idxlist = getattr(self, key).tolist()
            idxs.append([idxlist.index(i) for i in table[key].array])

        # Add the results to the result arrays
        for rvid in self.result_variables:
            arr = getattr(self, rvid)

            # If the result variable is in the dataframe
            if rvid in table.columns:
                arr[tuple(idxs)] = table[rvid].to_numpy()

            # Otherwise, translate tspec to tp, or tp to tspec
            else:
                if rvid == "tspec" and ("tp" in table.columns):
                    arr[tuple(idxs)] = table["tp"] / 1.1
                elif rvid == "tp" and ("tspec" in table.columns):
                    arr[tuple(idxs)] = table["tspec"] * 1.1
                else:
                    raise KeyError(rvid)

    def extend(
        self,
        input_variable: str,
        grid: Union[list, np.ndarray],
        include_bounds: bool = False,
        merge_grid: bool = True,
    ) -> None:
        """
        Extend the grid of the input variable and therefore also the output variables.

        Parameters
        ----------
        input_variable : str
            The name of the input variable.
        grid : Union[list, np.ndarray]
            The 1D grid values to which the input variable should be extended.
        include_bounds : bool, optional
            Whether or not to include values within the upper and lower bound of the input variable (default is False).
        merge_grid : bool, optional
            Whether or not to merge the new grid with the existing grid of the input variable (default is True).
        """
        # Haal de huidige discretisatie van de variabele op
        xp = getattr(self, input_variable)
        axis = self.input_variables.index(input_variable)

        # If include_bounds = True, add values between the min and max, otherwise add all
        if not include_bounds:
            x = np.array(
                [
                    val
                    for val in np.atleast_1d(grid)
                    if not (xp.min() <= val <= xp.max())
                ]
            )

        # If merge_grid, add the grid to the existing values
        if merge_grid:
            x = np.array(sorted(set(np.atleast_1d(grid)).union(xp)))

        # If x is empty or the requested values are equal to those already present, continue
        if not any(x) or np.array_equal(x, xp):
            return None

        # Extend all result variables
        intstr = InterpStruct(x=x, xp=xp)
        for resvar in self.result_variables:
            # Obtain the result variable
            arr = getattr(self, resvar)
            if np.isnan(arr).any():
                raise ValueError(
                    f'[ERROR] NaN values ({np.isnan(arr).sum()}) in array "{resvar}" to interpolate.'
                )

            # Interpolate wave conditions
            if resvar in ["hs", "tp", "tspec"]:
                arr = np.maximum(0.0, intstr.interp(fp=arr, axis=axis))

            # Interpoleer wave angle
            elif resvar == "dir":
                arr = intstr.interp_angle(fp=arr, axis=axis)

            # Use of VZM, (TODO find out: interpolate between 0 and 1?)
            elif resvar == "vzm":
                arr = intstr.interp(fp=arr, axis=axis)

            # Interpolate other values
            else:
                arr = intstr.interp(fp=arr, axis=axis)

            # Add the extended result variable array back to the object
            setattr(self, resvar, arr)

        # Add the extended input variable back to the object
        setattr(self, input_variable, x)

    def refine(
        self, result_variable: str, grid: Dict[str, Union[list, np.ndarray]]
    ) -> np.ndarray:
        """
        Extend the grid of the input variable and therefore also the output variables.

        Parameters
        ----------
        input_variable : str
            The name of the input variable.
        result_variable : str
            The name of the output variable.
        grid : Dict[str, Union[list, np.ndarray]]
            Input variables with their corresponding grid to extend the result_variable to

        Returns
        -------
        np.ndarray
            The adjusted grid.
        """
        # Controleer of variabele aanwezig is
        if result_variable not in self.result_variables:
            raise KeyError(
                f"[ERROR] Result variable '{result_variable}' not in loading model."
            )

        # Obtain the array from the result variable
        belasting_int = getattr(self, result_variable)

        # Loop over the different grid items
        for inpvar, x in grid.items():
            if inpvar not in self.input_variables:
                raise KeyError(
                    f"[ERROR] Input variable '{inpvar}' not in loading model ({', '.join(self.input_variables)})"
                )

            # Obtain the current grid and axis
            xp = getattr(self, inpvar)
            axis = self.input_variables.index(inpvar)

            # If x and xp are equal, no interpolation needed
            if np.array_equal(x, xp):
                continue

            # If all x in xp, just select the requested values
            if np.isin(x, xp).all():
                idx = np.isin(xp, x)
                belasting_int = np.take(
                    belasting_int, indices=np.where(idx)[0], axis=axis
                )
                continue

            # If xp is just one value, duplicate
            if len(xp) == 1:
                belasting_int = np.tile(belasting_int, (1, 1, len(x)))
                continue

            # Interpolate
            intstr = InterpStruct(x=x, xp=xp)

            # Interpolate wave conditions
            if result_variable in ["hs", "tp", "tspec"]:
                belasting_int = np.maximum(
                    0.0, intstr.interp(fp=belasting_int, axis=axis)
                )

            # Interpolate wave direction
            elif result_variable == "dir":
                belasting_int = intstr.interp_angle(fp=belasting_int, axis=axis)

            # Use of VZM, (TODO find out: interpolate between 0 and 1?)
            elif result_variable == "vzm":
                belasting_int = intstr.interp(fp=belasting_int, axis=axis)

            # Interpolate other values
            else:
                belasting_int = intstr.interp(fp=belasting_int, axis=axis)

        # Return the interpolated array for the result variable
        return belasting_int

    def repair(
        self,
        input_variable: str,
        result_variables: Union[str, list] = None,
        epsilon: float = 1e-6,
    ) -> None:
        """
        Make the result values of the given output variable monotonically increasing along the axis of a given input variable.

        Parameters
        ----------
        input_variable : str
            The name of the input variable along which the result values should be made monotonically increasing.
        result_variables : Union[str, list]
            The name of the output variable to be made monotonically increasing.
        epsilon : float, optional
            The minimum difference between the values of the repaired output variable (default is 1e-6).
        """
        # Obtain the relevant axis
        axis = self.input_variables.index(input_variable)

        # If no result variables are given, take all result variables in the model
        if result_variables is None:
            result_variables = self.result_variables

        # Loop over the result variables
        for var in np.atleast_1d(result_variables):
            # Obtain the result variable grid
            arr = getattr(self, var)
            rows = [np.take(arr, indices=0, axis=axis)]

            # Make monotonous increasing over the axis of the input_variable
            for i in range(1, arr.shape[axis]):
                last = rows[-1]
                nxtt = np.take(arr, indices=i, axis=axis)
                rows.append(np.maximum(last + epsilon, nxtt))

            # Add the adjusted grid to the loadingmodel
            setattr(self, var, np.stack(rows, axis=axis))

    def calculate_hbn(
        self, profile: Profile, qcrit: float, factor_hs: float, factor_tspec: float
    ):
        """
        Add hbn result variables to each of the LoadingModels.
        If 'hbn' is already defined, it will overwrite the old result variable.

        Parameters
        ----------
        profile : Profile
            The profile
        qcrit : float
            The critical discharge
        factor_hs : float
            Factor for the significant wave height, used for model uncertainty
        factor_tspec : float
            Factor for the spectral wave period, used for model uncertainty
        """
        # Controleer of de juiste resultaatwaarden aanwezig zijn
        for resvar in ["h", "hs", "tspec", "dir"]:
            if not hasattr(self, resvar):
                raise KeyError(
                    f"Resultaatvariabele '{resvar}' is nodig voor kruinhoogteberekening, maar niet aanwezig."
                )

        # Prepare wave conditions
        _h = self.h.ravel()
        _hs = np.array(self.hs * factor_hs).ravel()
        _tspec = np.array(self.tspec * factor_tspec).ravel()
        _dir = np.array(self.dir).ravel()

        # Calculate HBN
        self.hbn = np.reshape(
            [profile.calculate_crest_level(qcrit, _h, _hs, _tspec, _dir)], self.h.shape
        )

        # Add the result variable
        self.result_variables.append("hbn")



import numpy as np

from .calculation import Calculation
from .datamodels.frequency_line import FrequencyLine
from ..location.location import Location


class ExceedanceFrequencyLineExperimental(Calculation):
    """
    Calculate a frequency line for a result variable (e.g. h (waterlevel), hs (significant wave height)) for a location
    """

    def __init__(
        self,
        result_variable: str,
        model_uncertainty: bool = True,
        levels: list = None,
        step_size: float = 0.1,
    ):
        """
        The __init__ method initializes an instance of the ExceedanceFrequencyLine class. It takes in several parameters to configure the calculation of the frequency line.

        Parameters
        ----------
        result_variable : str
            The result variable for which the frequency line will be calculated.
        model_uncertainty : bool (optional)
            Enable or disable the use of model uncertainties when calculating the frequency line. Default is True.
        levels : list (optional):
            The levels at which the exceedance probability has to be calculated. If not specified, the levels will be chosen between the 1st and 99th percentile of the values in the HRDatabase.
        step_size : float (optional)
            The step size of the frequency line. Default is 0.1.
        """
        # Inherit
        super().__init__()

        # Warning
        print(
            "Experimental version may give different results for water levels when applying model uncertainty."
        )

        # Save settings
        self.set_result_variable(result_variable.lower())
        self.use_model_uncertainty(model_uncertainty)
        self.set_levels(levels)
        self.set_step_size(step_size)

    def calculate_location(self, location: Location) -> FrequencyLine:
        """
        Calculate the exceedance probability of the variable at a given set of levels.
        If the levels are not specified, they will be chosen at the 1st and 99th percentile of all values in the database.

        Parameter
        ---------
        location : Location
            The Location object

        Returns
        -------
        FrequencyLine
            Frequency line of the result variable
        """
        # Copy the levels
        levels = self.levels

        # Obtain location object
        model = location.get_model()
        loading = model.get_loading()

        # Check if the levels are defined, if not, define it between the 1st and 99th percentile
        if levels is None:
            lower, upper = loading.get_quantile_range(
                self.result_variable, 0.01, 0.99, 3
            )
            levels = np.arange(lower, upper + 0.5 * self.step_size, self.step_size)

        # Repair
        loading.repair_loadingmodels(self.result_variable)

        # Splits uit naar trage stochasten en windrichting
        p_h_slow = model.calculate_probability_loading(
            result_variable=self.result_variable,
            levels=levels,
            model_uncertainty=self.model_uncertainty,
            split_input_variables=list(model.statistics.stochastics_slow.keys()),
            given=list(model.statistics.stochastics_slow.keys()),
        )

        # Reken kansen om naar overschrijdingskansen door over de eerste te sommeren
        ep_h_slow = np.cumsum(p_h_slow[::-1], axis=0)[-2::-1]

        # Process slow stochastics (they are always at the last axes of the matrix)
        if len(list(model.statistics.stochastics_slow.keys())) > 0:
            p_trapezoidal = model.process_slow_stochastics(ep_h_slow)
            exceedance_probability = (
                p_trapezoidal * location.get_settings().periods_base_duration
            )

        # Zo niet, geef de overschrijdingskansen direct terug
        else:
            exceedance_probability = (
                ep_h_slow * location.settings.periods_block_duration
            )

        # Return the frequency line
        return FrequencyLine(levels, exceedance_probability)

    def set_result_variable(self, result_variable: str):
        """
        Change the result variable for which the frequency line will be calculated.

        Parameters
        ----------
        result_variable : str
            The result variable for which the frequency line will be calculated
        """
        # Raise an error when assigning the wave direction (dir)
        if result_variable == "dir":
            raise ValueError(
                "[ERROR] Cannot calculate a frequency line for the wave direction (dir)."
            )

        # Save result variable
        self.result_variable = result_variable

    def set_levels(self, levels: list = None):
        """
        Change the levels.
        If levels is not defined, the frequency line is calculated based upon the 1st and 99th percentile.

        Parameters
        ----------
        levels : list, optional
            The levels at which the exceedance probability has to be calculated
        """
        self.levels = levels

    def set_step_size(self, step_size: float):
        """
        Change the step size of the frequency line.

        Parameters
        ----------
        step_size : float
            The step size of the frequency line
        """
        # Cannot be smaller or equal to 0
        if step_size <= 0:
            raise ValueError("[ERROR] Step size should be larger than 0.")

        # Save step size
        self.step_size = step_size

    def use_model_uncertainty(self, model_uncertainty: bool):
        """
        Use model uncertainty when calculating a frequency line.

        Parameters
        ----------
        model_uncertainty : bool
            Enable or disable the use of model uncertainties
        """
        self.model_uncertainty = model_uncertainty



from .loading import Loading
from .water_systems.loading_coast import LoadingCoast
from .water_systems.loading_eastern_scheldt import LoadingEasternScheldt
from .water_systems.loading_ijssel_vechtdelta import LoadingIJsselVechtdelta
from .water_systems.loading_lake import LoadingLake
from .water_systems.loading_lower_rivier import LoadingLowerRiver
from .water_systems.loading_upper_river import LoadingUpperRiver
from ...settings.settings import Settings
from ....common.enum import WaterSystem


class LoadingFactory:
    """
    A factory class to generate the right Loading object for a given Settings object.

    Attributes
    ----------
    WATER_SYSTEM_LOADING_MAP : dict
        A dictionary containing the corresponding Loading class for a WaterSystem
    """

    # Dictionary with Loading classes for each WaterSystem
    WATER_SYSTEM_LOADING_MAP = {
        # Upper River
        WaterSystem.RHINE_NON_TIDAL: LoadingUpperRiver,
        WaterSystem.MEUSE_NON_TIDAL: LoadingUpperRiver,
        WaterSystem.MEUSE_VALLEY_NON_TIDAL: LoadingUpperRiver,
        # Lower River
        WaterSystem.RHINE_TIDAL: LoadingLowerRiver,
        WaterSystem.MEUSE_TIDAL: LoadingLowerRiver,
        # TODO: WaterSystem.EUROPOORT
        # Coast
        WaterSystem.WADDEN_SEA_EAST: LoadingCoast,
        WaterSystem.WADDEN_SEA_WEST: LoadingCoast,
        WaterSystem.COAST_NORTH: LoadingCoast,
        WaterSystem.COAST_CENTRAL: LoadingCoast,
        WaterSystem.COAST_SOUTH: LoadingCoast,
        WaterSystem.WESTERN_SCHELDT: LoadingCoast,
        # Eastern Scheldt
        WaterSystem.EASTERN_SCHELDT: LoadingEasternScheldt,
        # Lakes
        WaterSystem.IJSSEL_LAKE: LoadingLake,
        WaterSystem.MARKER_LAKE: LoadingLake,
        # TODO: WaterSystem.VELUWE_LAKES
        # TODO: WaterSystem.GREVELINGEN
        # IJssel-Vecht Delta
        WaterSystem.VECHT_DELTA: LoadingIJsselVechtdelta,
        WaterSystem.IJSSEL_DELTA: LoadingIJsselVechtdelta,
        # Volkerak-Zoommeer
        # TODO: WaterSystem.VOLKERAK_ZOOMMEER
        # Hollandsche IJssel
        # TODO: WaterSystem.HOLLANDSCHE_IJSSEL
        # Other
        # WaterSystem.COAST_DUNES
        # WaterSystem.DIEFDIJK
    }

    @staticmethod
    def get_loading(settings: Settings) -> Loading:
        """
        Returns the Loading object for the specified water system.

        Parameters
        ----------
        settings : Settings
            A Settings object to get the loading for.

        Returns
        -------
        Loading
            A Loading object for the specified water system.

        Raises
        ------
        NotImplementedError
            If loading for the specified water system are not implemented.
        """
        # Obtain the right Loading class
        loading_class = LoadingFactory.WATER_SYSTEM_LOADING_MAP.get(
            settings.watersystem
        )

        # Return if the class is found, otherwise raise an error
        if loading_class:
            return loading_class(settings)
        else:
            raise NotImplementedError(
                f"[ERROR] Loading for '{settings.watersystem.name}' not implemented (ID: {settings.watersystem.value})"
            )



import numpy as np
import os


class FileHydraNL:
    """
    Class with staticsmethods to read Hydra-NL input files
    """

    # Path to the package
    PACKAGE_PATH = os.path.join(
        os.path.split(os.path.dirname(__file__))[0], "data", "statistics"
    )

    @staticmethod
    def read_file_2columns(path):
        """
        Read a Hydra-NL statistics file with two columns

        Parameters
        ----------
        path : str
            Path to the statistics file
        """
        # Open file
        if "\\" in path:
            path = path.replace("\\", "/")
        with open(
            os.path.join(FileHydraNL.PACKAGE_PATH, path), "r", encoding="cp1252"
        ) as f:
            lines = f.readlines()

        # Read non commented values from file, and split per line
        vals = [line.strip().split() for line in lines if not line.startswith("*")]

        # Convert to float and split in columns
        kol1, kol2 = np.vstack(vals).astype(float).T

        return kol1, kol2

    @staticmethod
    def read_file_ncolumns(path):
        """
        Read a Hydra-NL statistics file with more than two columns

        Parameters
        ----------
        path : str
            Path to the statistics file
        """
        # Open file
        if "\\" in path:
            path = path.replace("\\", "/")
        with open(
            os.path.join(FileHydraNL.PACKAGE_PATH, path), "r", encoding="cp1252"
        ) as f:
            lines = f.readlines()

        # Read non commented values from file, and split per line
        vals = [line.strip().split() for line in lines if not line.startswith("*")]

        # Convert to float and split in columns
        floatvals = np.vstack(vals).astype(float)

        return floatvals[:, 0], floatvals[:, 1:]

    @staticmethod
    def read_file_ncolumns_loc(path):
        """
        Read a Hydra-NL statistics location file with more than two columns

        Parameters
        ----------
        path : str
            Path to the statistics file
        """
        # Open file
        if "\\" in path:
            path = path.replace("\\", "/")
        with open(
            os.path.join(FileHydraNL.PACKAGE_PATH, path), "r", encoding="cp1252"
        ) as f:
            lines = f.readlines()

        # Process data
        vals_tot = [line.strip().split() for line in lines]

        # Convert to float and split in columns
        x = int(vals_tot[2][1][:-1])
        y = int(vals_tot[2][2])

        return x, y



import numpy as np

# from scipy.stats import norm

from .barrier import Barrier

# from ..discharge import Discharge
from ..discrete_probability import DiscreteProbability

# from ..sea_level.sea_level import SeaLevel
# from ..wind_speed import WindSpeed
# from ....loading.loading_model.loading_model import LoadingModel
from .....settings.settings import Settings


class BarrierRamspol(Barrier):
    def __init__(self, settings: Settings, wind_direction: DiscreteProbability):
        # Inherit
        super().__init__(settings, wind_direction)

        # Init standard variables
        self.k = [1, 2]
        self.nk = len(self.k)

        # Save failure probability of the Ramspol
        self.failure_probability_ramspol = settings.failure_probability_ramspol

    def calculate_closing_probability(
        self, wind_direction: float, closing_situation: int
    ) -> np.ndarray:
        """
        Calculate the failure probability of the Ramspol
        """
        if closing_situation == 1:
            return self.failure_probability_ramspol
        elif closing_situation == 2:
            return 1 - self.failure_probability_ramspol
        else:
            raise KeyError(
                f"[ERROR] Unknown closing situation: {closing_situation}, expecting 1 or 2."
            )



import numpy as np

from abc import ABC, abstractmethod
from typing import Tuple, Union

from ...settings.settings import Settings
from ....common.interpolate import InterpStruct
# from ....io.database_hr import DatabaseHR


class Loading(ABC):
    """
    Loading Abstract Base Class. This class contains all LoadingModels.

    Attributes
    ----------
    settings : Settings
        The Settings object
    model : dict
        Dictionary with all LoadingModels (wind direction, closing situation)
    database : DatabaseHR
        Connection with HR Database
    """

    def __init__(self, settings: Settings):
        """
        Init the Loading object

        Parameters
        ----------
        settings : Settings
            The Settings object
        """
        # Save settings and init dictionary for all models
        self.settings = settings
        self.model = {}

    @abstractmethod
    def read_loading(self) -> None:
        """
        Read the HR and create Loading Models
        """
        pass

    def iter_models(self):
        """
        Iterate through each of the LoadingModels.
        LoadingModels are ordered by WindDirection, ClosingSituation
        """
        for discreet, model in self.model.items():
            yield discreet, model

    def _extend_loadingmodels(self) -> None:
        """
        Preprocess the LoadingModels.
        Makes sure the discretisations in every LoadingModel are the same.
        """
        # Obtain the wind directions and closing situations
        self.r = sorted(set([r[0] for r in list(self.model.keys())]))
        self.k = sorted(set([k[1] for k in list(self.model.keys())]))

        # Collect all unique wind speeds from all loadingmodels
        utot = sorted(set([u for _, model in self.iter_models() for u in model.u]))

        # Extend the stochasts in each loadingmodel
        for _, model in self.iter_models():
            # Extend the wind speed
            model.extend("u", utot)

            # Extend for discharge
            if "q" in model.input_variables:
                if len(model.q) > 1:
                    model.extend(
                        "q",
                        list(filter(None, [self.settings.q_min, self.settings.q_max])),
                    )

            # Extend for sea level
            if "m" in model.input_variables:
                if len(model.m) > 1:
                    model.extend(
                        "m",
                        list(filter(None, [self.settings.m_min, self.settings.m_max])),
                    )

            # Extend for lake level
            if "a" in model.input_variables:
                if len(model.a) > 1:
                    model.extend(
                        "a",
                        list(filter(None, [self.settings.a_min, self.settings.a_max])),
                    )

    def repair_loadingmodels(
        self, result_variables: Union[list, str], epsilon: float = 1e-6
    ) -> None:
        """
        Repair the result variables for all LoadingModels
        Repairs depending on the '{input_variable}_repair' flag in the Settings

        Parameters
        ----------
        result_variables : Union[list, str]
            Result variables
        epsilon : float
            The minimum increase when repairing (default : 1e-6)
        """
        # For each LoadingModel (wind direction and closing situation)
        for _, model in self.iter_models():
            # Loop over the inputvariables
            for inpvar in model.input_variables:
                # Only repair base stochastics
                if inpvar not in ["u", "m", "q", "a"]:
                    continue

                # Only if the flag '{input_variable}_repair' is True
                if not getattr(self.settings, f"{inpvar}_repair"):
                    continue

                # For each result variable
                for rv in np.atleast_1d(result_variables):
                    # If the result variable from the argument is not in the LoadingModel or if it the wave direction, skip
                    if (rv not in model.result_variables) or (rv == "dir"):
                        continue

                    # Repair the LoadingModel
                    model.repair(inpvar, rv, epsilon)

    def get_result_variable_statistic(self, result_variable: str, stat: str, args=()):
        """
        Obtain statistics for a result variable. (e.g. min, max)

        Parameters
        ----------
        result_variable : str
            Result variable (h, hs, tp, tspec, hbn, qov, etc.)
        statistic : str
            NumPy function, e.g. 'max', 'min'
        args : tuple
            Optional if required by NumPy function

        Returns
        -------
        statistic : float
            The statistic
        """
        # Obtain raveled result variable array
        arr = self.get_result_variable_raveled(result_variable)

        # Obtain the statistics using the raveled result variable array
        result = getattr(np, stat)(arr, *args)

        # Return statistic
        return result

    def get_result_variable_raveled(self, result_variable: str) -> np.ndarray:
        """
        Obtain all values for a result variable in one array

        Parameters
        ----------
        result_variable : str
            Result variable (h, hs, tp, tspec, hbn, qov, etc.)

        Returns
        -------
        np.ndarray
            1D array with all values of a certain result variable
        """
        # Loop over every discrete model and add all result variables values to an array
        arr = np.concatenate(
            [
                getattr(model, result_variable).ravel()
                for _, model in self.iter_models()
            ],
            axis=-1,
        )

        # Return the array
        return arr

    def get_quantile_range(
        self,
        result_variable: str,
        lower_quantile: float,
        upper_quantile: float,
        round_digits: int = None,
    ) -> Tuple[float, float]:
        """Geef kwantielen van een gekozen variabele. De uitkomst kan worden
        afgerond wanneer 'rounddigits' is gegeven.

        Parameters
        ----------
        result_variable : str
            Result variable (h, hs, tp, tspec, hbn, qov, etc.)
        lower_quantile : float
            Lower quantile, between 0.0 en 1.0
        upper_quantile : float
            Upper quantile, between 0.0 en 1.0
        round_digits : int, optional
            Aantal digits waarop afgerond wordt, standaard None

        Returns
        -------
        tuple
            Parameterwaarden voor het onder en bovenkwantiel.
        """
        # Obtain all values of the result variable in the discrete models in one array
        arr = self.get_result_variable_raveled(result_variable)

        # Determine lower and upper quantile
        lower = np.quantile(arr, lower_quantile)
        upper = np.quantile(arr, upper_quantile)

        # If required, apply rounding
        if round_digits is not None:
            lower = np.round(lower, round_digits)
            upper = np.round(upper, round_digits)

        # Return the quantiles
        return lower, upper

    def get_wave_conditions(
        self,
        direction: float,
        waterlevel: Union[int, float, list, np.ndarray],
        extrapolate: bool = True,
    ) -> Tuple[dict, np.ndarray]:
        """
        Return the wave conditions given a wind direction and wind speed.

        The function assumes the wave conditions are a function of:
        - The local water level
        - The wind direction
        - The wind speed

        Parameters
        ----------
        direction : float
            The wind direction
        waterlevel : Union[float, list, np.ndarray]
            Water level(s)
        extrapolate : bool, optional
            Whether or not to extrapolate (default : True)

        Returns
        -------
        Tuple[dict{str : np.ndarray}, np.ndarray]
            1.) A dictionary with
                key : The loading variable (e.g. h, hs, tspec, tp, dir)
                np.ndarray : 2D array with size [u, waterlevel]
            The wave conditions for different wind speed (nd-array)
            2.) The wind speed discretistation (1d-array)
        """
        # Convert the water level to a ndarray
        if isinstance(waterlevel, (float, int)):
            waterlevel = np.array([waterlevel])
        elif isinstance(waterlevel, list):
            waterlevel = np.array(waterlevel)

        # Result variables
        resvars = ["h", "hs", "tspec", "tp", "dir"]

        # Obtain all loading models for one wind direction
        loading_r = [model for key, model in self.model.items() if direction in key]
        first_model = loading_r[0]

        # Create an empty dictionary for the wave conditions
        wave_conditions = {
            resvar: np.zeros((len(waterlevel), len(first_model.u)), dtype=np.float32)
            for resvar in resvars
        }

        # Change the loading shape such that all wave conditions are available per wind direction and wind speed
        loading_reshaped = {}

        # For each loading parameter (resvar)
        for resvar in resvars:
            # Obtain the loading array for each loading model, stack over the last dimension such
            # that the input order of the other stochastics are unchanged: e.g. parts[model, u, m, d, p]
            parts = []
            for model in loading_r:
                if hasattr(model, resvar):
                    parts.append(getattr(model, resvar))
                else:
                    if (resvar == "tp") and hasattr(model, "tspec"):
                        parts.append(getattr(model, "tspec") * 1.1)
                    elif (resvar == "tspec") and hasattr(model, "tp"):
                        parts.append(getattr(model, "tp") / 1.1)
                    else:
                        raise KeyError(f'"{resvar}" not present.')

            # e.g. arr[u, m, d, p, model]
            arr = np.stack(parts, axis=-1)

            # Reshape
            upos = first_model.input_variables.index("u")

            # If the axis of the wind speed is not the first dimension, swap axes
            if upos != 0:
                # Swap to axis with the wind speed to the first dimension
                tmparr = np.swapaxes(arr, 0, upos)

                # Reshaped [u, comb van (m, d, p en de sluitsituatie (model))]
                reshaped = tmparr.reshape((tmparr.shape[0], np.prod(tmparr.shape[1:])))

            # Otherwise, only reshape the array
            else:
                reshaped = arr.reshape((arr.shape[0], np.prod(arr.shape[1:])))

            # Determine the order of the water levels
            if resvar == "h":
                order = np.argsort(reshaped, axis=1)

            # Sort
            i = np.arange(reshaped.shape[0])[:, None]
            loading_reshaped[resvar] = reshaped[i, order]

        # Obtain the waterlevels, over which will be interpolated
        hbelast = loading_reshaped["h"]

        # Interpolate over the water level, for the wind speed
        for iu in range(len(first_model.u)):
            # Prepare interpolation
            index = np.unique(hbelast[iu].round(3), return_index=True)[1]
            h_unique = hbelast[iu][index]

            # If there is only one water level given, we dont need to interpolate
            if len(h_unique) == 1:
                for resvar in resvars:
                    wave_conditions[resvar][:, iu] = loading_reshaped[resvar][iu][index]

            # For more water levels, we do need to interpolate
            else:
                intstr = InterpStruct(x=waterlevel, xp=h_unique)
                for resvar in resvars:
                    # Dont interpolate the water level, we will add this at the end
                    if resvar == "h":
                        continue
                    fp = loading_reshaped[resvar][iu][index]

                    # If all result variables are 0, skip
                    if (fp == 0.0).all():
                        continue

                    # If there is only one result variable, apply this one
                    if (fp == fp[0]).all():
                        intbelast = fp[0]

                    # Otherwise, interpolate the wave loading
                    else:
                        if resvar == "dir":
                            intbelast = intstr.interp_angle(
                                fp=fp, extrapolate=extrapolate
                            )
                        else:
                            intbelast = np.maximum(
                                0, intstr.interp(fp=fp, extrapolate=extrapolate)
                            )

                    # Add the resvar to the result 'wave_conditions' dictionary
                    wave_conditions[resvar][:, iu] = intbelast

        # Add the water level
        wave_conditions["h"][:, :] = waterlevel[:, None]

        # Return wave conditions and the wind speed discretisation
        return wave_conditions, first_model.u



from typing import List

from ..common.enum import WaterSystem
from ..io.database_hr import DatabaseHR
from ..location.location import Location
from ..location.settings.settings import Settings


class HRDatabase:
    """
    HRDatabase class

    Attributes
    ----------
    watersystem, WaterSystem
        Water system of the dike trajectory
    locations, list
        List with all locations in the database
    """

    def __init__(self, database_path: str) -> None:
        # Connect to the database
        self.database_path = database_path

        # Obtain water system and locations
        with DatabaseHR(database_path) as database:
            self.watersystem = database.get_water_system()
            self.locationnames = database.get_hrdlocations_names()

        # Empty locations
        self.locations = {}

    def __len__(self) -> int:
        """
        Return the number of locations

        Returns
        -------
        int
            Number of parts
        """
        return len(self.locationnames)

    def __iter__(self):
        """
        Dunder to allow iterating through the locations
        """
        self.__iterindex = -1
        return self

    def __next__(self):
        """
        Dunder to allow iterating through the locations
        """
        self.__iterindex += 1
        if self.__iterindex < len(self.locationnames):
            return self.locationnames[self.__iterindex]
        else:
            raise StopIteration

    def get_settings(self, hrdlocation: str) -> Settings:
        """
        Returns the Settings class of a hrdlocation
        Useful for when you manually want to adjust the settings
        To create a location using the settings use the function 'create_location(settings)'

        Parameters
        ----------
        hrdlocation : str
            The hrdlocation (uitvoerpunt)

        Returns
        -------
        Settings
            The Settings object for the hrdlocation
        """
        # Check if the hrdlocation exists
        if self.check_location(hrdlocation):
            # Return the settings
            return Settings(hrdlocation, self.database_path)

        # Otherwise, raise an exception
        else:
            raise ValueError(
                f"[ERROR] HRDLocation '{hrdlocation}' not found in the database."
            )

    def create_location(self, settings: Settings) -> Location:
        """
        Creates a location based upon the Settings object

        Parameters
        ----------
        settings : Settings
            The Settings object of a location

        Returns
        -------
        Location
            The Location object for the given settings
        """
        # Create the location and save it in the cache
        self.locations[settings.location] = Location(settings)

        # Return location
        return self.locations[settings.location]

    def check_location(self, hrdlocation: str) -> bool:
        """
        Check if a hrdlocation exists within the database

        Parameters
        ----------
        hrdlocation : str
            The hrdlocation (uitvoerpunt)

        Returns
        -------
        bool
            Whether the hrdlocation is in the database
        """
        return hrdlocation in self.locationnames

    def get_location(self, hrdlocation: str) -> Location:
        """
        Returns the Location class of a hrdlocation.
        Uses the default set of parameters.

        Parameters
        ----------
        hrdlocation : str
            The hrdlocation (uitvoerpunt)

        Returns
        -------
        Location
            The Location object for the hrdlocation
        """
        # Check if the hrdlocation exists
        if self.check_location(hrdlocation):
            # If the location is not cached, create the location
            if hrdlocation not in self.locations:
                # Default settings and create location
                settings = self.get_settings(hrdlocation)
                return self.create_location(settings)

            # Return the cached location
            else:
                return self.locations[hrdlocation]

        # Otherwise, raise an exception
        else:
            raise ValueError(
                f"[ERROR] HRDLocation '{hrdlocation}' not found in the database."
            )

    def get_location_names(self) -> List[str]:
        """
        Return a list of all hrdlocations in the database

        Returns
        -------
        list[str]
            List with all hrdlocation names
        """
        return self.locationnames

    def get_water_system(self) -> WaterSystem:
        """
        A method that returns the water system corresponding to the trajectory.

        Returns
        -------
        WaterSystem
            Water system of the trajectory
        """
        return self.watersystem



import numpy as np

from .base_model import BaseModel
from .loading.other_systems.loading_wave_overtopping import LoadingWaveOvertopping
from .statistics.other_systems.statistics_wave_overtopping import (
    StatisticsWaveOvertopping,
)
from ..location import Location


class WaveOvertopping(BaseModel):
    def __init__(
        self, location: Location, water_levels: np.ndarray, probability: np.ndarray
    ):
        """
        Wave Overtopping model
        """
        # Inherit
        super().__init__(location.get_settings())

        # Init statistics
        self.statistics = StatisticsWaveOvertopping(location, water_levels, probability)

        # Init loading
        self.loading = LoadingWaveOvertopping(location, water_levels)



__version__ = "0.0.4"
#############################################################
# Pydra (v2)
# Contact: n.vandervegt@hkv.nl
#############################################################

# Core
from .core.exceedance_frequency_line import ExceedanceFrequencyLine
from .core.exceedance_frequency_line_experimental import (
    ExceedanceFrequencyLineExperimental,
)
from .core.hbn import HBN

# Enums
from .common.enum import Breakwater
from .common.enum import WaterSystem

# Other
from .hrdatabase.hrdatabase import HRDatabase
from .location.location import Location
from .location.profile.profile import Profile

__all__ = [
    "ExceedanceFrequencyLine",
    "ExceedanceFrequencyLineExperimental",
    "HBN",
    "Breakwater",
    "WaterSystem",
    "HRDatabase",
    "Location",
    "Profile",
]



from ....settings.settings import Settings
from .....io.file_hydranl import FileHydraNL


class SigmaFunction:
    """
    Class for the sigma function, which describes the correlation between the
    sea level and wind speed.
    """

    def __init__(self, settings: Settings):
        """
        Constructor class for the SigmaFunction.

        Parameters
        ----------
        settings : Settings
            The Settings object
        """
        # Read the statistics file
        self.sigma_sea_level, self.sigma = FileHydraNL.read_file_ncolumns(
            settings.sigma_function
        )
        self.correlation = self.sigma.min(axis=0) > 0.0



import numpy as np

from .barrier import Barrier


class NoBarrier(Barrier):
    def __init__(self):
        # Inherit
        super().__init__(None, None)

        # Only 1 barrier state (= open)
        self.k = [1]
        self.nk = len(self.k)

    def calculate_closing_probability(
        self, wind_direction: float, closing_situation: int
    ) -> np.ndarray:
        """
        Irrelevant for no barrier
        """
        return None



import numpy as np

from abc import ABC


class SeaLevel(ABC):
    """
    Class to describe the sea (water) level statistics.
    Sea level statistics are conditional on the wind direction.
    """

    def __init__(self):
        """
        Constructor class for the SeaLevel statistics.
        """
        self.nm = 0
        self.m = None
        self.epm = None

    def __len__(self):
        """
        Return the number of sea level discretisations.

        Returns
        -------
        int
            Number of discretisations
        """
        return self.nm

    def get_discretisation(self) -> np.ndarray:
        """
        Return the sea level discretisation.

        Returns
        -------
        np.ndarray
            1D array with discretisation
        """
        return self.m

    def get_exceedance_probability(self) -> np.ndarray:
        """
        Return exceedance probility of the sea level, conditional on the wind
        direction.

        Returns
        -------
        np.ndarray
            A 2D array with the sea level exceedance probability conditional on
            the wind direction (r x m)
        """
        return self.epm



import numpy as np
import pandas as pd

from itertools import product
from scipy.interpolate import interp1d

from ..loading import Loading
from ..loading_model.loading_model import LoadingModel
from ....settings.settings import Settings
from .....common.interpolate import InterpStruct
from .....io.database_hr import DatabaseHR


class LoadingEasternScheldt(Loading):
    """
    Loading class for the EasternScheldt
    Water systems: Eastern Scheldt
    """

    def __init__(self, settings: Settings):
        """
        Init the Loading object for the Eastern Scheldt

        Parameters
        ----------
        settings : Settings
            The Settings object
        """
        # Inherit the from parent
        super().__init__(settings)

        # Read and process the loading
        self.read_loading()

    def read_loading(self) -> None:
        """
        Read the HR result table and create LoadingModels
        """
        # Read table
        with DatabaseHR(self.settings.database_path) as database:
            waterlevels, waveconditions = database.get_result_table_eastern_scheldt(
                self.settings
            )
            ivids = database.get_input_variables()
            rvids = database.get_result_variables()

        # Remove the 'm_os' from the ivids (two times m/m_os because m is used in the wave conditions)
        ivids.remove("m_os")

        # The water level and wave conditions are based on different grids
        # Repair water level
        for richting, snelheid in product(
            np.unique(waveconditions["r"]), np.unique(waveconditions["u"])
        ):
            # Kijk of deze combi bestaat
            if (
                waterlevels["r"].eq(richting) & waterlevels["u"].eq(snelheid)
            ).sum() == 0:
                _format = waterlevels[
                    waterlevels["r"].eq(richting) & waterlevels["u"].eq(0)
                ].copy()
                _format.loc[:, "u"] = snelheid
                _format.loc[:, "h"] = np.NaN
                waterlevels = pd.concat([waterlevels, _format], ignore_index=True)

        # Repareer Hs
        for richting, snelheid in product(
            np.unique(waterlevels["r"]), np.unique(waterlevels["u"])
        ):
            # Kijk of deze combi bestaat
            if (
                waveconditions["r"].eq(richting) & waveconditions["u"].eq(snelheid)
            ).sum() == 0:
                _format = waveconditions[
                    waveconditions["r"].eq(richting) & waveconditions["u"].eq(0)
                ].copy()
                _format.loc[:, "u"] = snelheid
                _format.loc[:, "hs"] = np.NaN
                _format.loc[:, "tp"] = np.NaN
                _format.loc[:, "tspec"] = np.NaN
                _format.loc[:, "dir"] = np.NaN
                waveconditions = pd.concat([waveconditions, _format], ignore_index=True)

        # Sorteer
        waterlevels = waterlevels.sort_values(by=["k", "r", "m", "d", "p", "u"])
        waterlevels = waterlevels.reset_index(drop=True)
        waveconditions = waveconditions.sort_values(by=["r", "h", "u"])
        waveconditions = waveconditions.reset_index(drop=True)

        # Fix water levels
        missing_h_index = waterlevels[waterlevels["h"].isna()].index
        waterlevels.loc[missing_h_index, "h"] = np.around(
            waterlevels.loc[missing_h_index - 1, "h"].to_numpy()
            + (
                waterlevels.loc[missing_h_index, "u"].to_numpy()
                - waterlevels.loc[missing_h_index - 1, "u"].to_numpy()
            )
            * (
                (
                    waterlevels.loc[missing_h_index + 1, "h"].to_numpy()
                    - waterlevels.loc[missing_h_index - 1, "h"].to_numpy()
                )
                / (
                    waterlevels.loc[missing_h_index + 1, "u"].to_numpy()
                    - waterlevels.loc[missing_h_index - 1, "u"].to_numpy()
                )
            ),
            4,
        )

        # Vul de NaNs golfcondities
        for n, row in waveconditions[waveconditions["hs"].isna()].iterrows():
            wd = row["r"]
            wl = row["h"]
            ws = row["u"]
            grid = waveconditions[
                waveconditions["r"].eq(wd) & waveconditions["h"].eq(wl)
            ].dropna()
            waveconditions.loc[n, "hs"] = round(
                float(interp1d(grid["u"], grid["hs"], fill_value="extrapolate")(ws)), 4
            )
            waveconditions.loc[n, "tp"] = round(
                float(interp1d(grid["u"], grid["tp"], fill_value="extrapolate")(ws)), 4
            )
            waveconditions.loc[n, "tspec"] = round(
                float(interp1d(grid["u"], grid["tspec"], fill_value="extrapolate")(ws)),
                4,
            )

            # Niet de wave direction extrapoleren
            if np.min(grid["u"]) <= ws and ws <= np.max(grid["u"]):
                waveconditions.loc[n, "dir"] = round(
                    float(interp1d(grid["u"], grid["dir"])(ws)), 4
                )
            elif ws > np.max(grid["u"]):
                waveconditions.loc[n, "dir"] = grid["dir"].to_numpy()[-1]
            elif ws < np.min(grid["u"]):
                waveconditions.loc[n, "dir"] = grid["dir"].to_numpy()[0]

        # Interpoleer de golfcondities op de waterstanden, alvorens ze aan de belastingmodellen toe te voegen
        for (richting, snelheid), wavecond_ur in waveconditions.groupby(["r", "u"]):
            # Maak een index voor de selectie van windrichting en windsnelheid
            idx = waterlevels["r"].eq(richting) & waterlevels["u"].eq(snelheid)

            # Maak een interpolatiestructuur voor de windrichting en windsnelheid
            intstr = InterpStruct(
                x=waterlevels.loc[idx, "h"].to_numpy(), xp=wavecond_ur["h"].to_numpy()
            )

            # Interpoleer voor elk van de golfparameters
            for param in ["hs", "tp", "tspec", "dir"]:
                waterlevels.loc[idx, param] = intstr.interp(
                    fp=wavecond_ur[param].to_numpy()
                )

        # Init LoadingModels for each combination of wind direction (r) and closing situation (k)
        for comb, deeltabel in waterlevels.groupby(["r", "k"]):
            direction, closing_situation = comb

            # Create a LoadingModel
            model = LoadingModel(direction, closing_situation, ivids, rvids)
            model.initialise(deeltabel.copy())

            # Add model to the models dictionary
            self.model[comb] = model

        # Breidt het belastingmodel uit op basis van de invoergegevens
        self._extend_loadingmodels()
        self.repair_loadingmodels(rvids)



import numpy as np

from datetime import datetime

from .calculation import Calculation
from .datamodels.frequency_line import FrequencyLine
from ..common.probability import ProbabilityFunctions
from ..location.location import Location
from ..location.model.wave_overtopping import WaveOvertopping


class HBN(Calculation):
    """
    Calculate the HBN for a location
    """

    def __init__(
        self,
        q_overtopping: float = 0.010,
        levels: list = None,
        step_size: float = 0.1,
        model_uncertainty: bool = True,
    ):
        """
        The __init__ method initializes an instance of the ExceedanceFrequencyLine class. It takes in several parameters to configure the calculation of the frequency line.

        Parameters
        ----------
        q_overtopping : float (optional)
            The critical overtopping discharge. Default is 0.01 m3/m/s (10 l/m/s)
        levels : list (optional):
            The levels at which the exceedance probability has to be calculated. If not specified, the levels will be chosen between the 1st and 99th percentile of the values in the HRDatabase.
        step_size : float (optional)
            The step size of the frequency line. Default is 0.1.
        model_uncertainty : bool (optional)
            Enable or disable the use of model uncertainties when calculating the frequency line. Default is True.
        """
        # Inherit
        super().__init__()

        # Save settings
        self.set_critical_overtopping(q_overtopping)
        self.set_levels(levels)
        self.set_step_size(step_size)
        self.use_model_uncertainty(model_uncertainty)

    def calculate_location(self, location: Location) -> float:
        """
        Calculate the exceedance probability of the variable at a given set of levels.
        If the levels are not specified, they will be chosen at the 1st and 99th percentile of all values in the database.

        Parameter
        ---------
        location : Location
            The Location object

        Returns
        -------
        FrequencyLine
            Frequency line of the crest level
        """
        # Check if a profile is defined
        if not location.has_profile():
            print(
                f"[WARNING] Profile is not assiged to location '{location.get_settings().location}'. Skipping this calculation."
            )
            return None

        # Obtain profile and validate
        profile = location.get_profile()
        if not profile.validate_profile():
            print(
                f"[WARNING] Something is wrong with the assigned profile of location '{location.get_settings().location}'. Skipping this calculation."
            )
            return None

        # Copy the levels
        levels = self.levels

        # Obtain location object
        settings = location.get_settings()
        model = location.get_model()
        statistics = model.get_statistics()
        loading = model.get_loading()
        slow_stochastics = list(statistics.stochastics_slow.keys())

        # TODO: Beter levels bepalen
        # If no levels are defined, derive them based on the water level in the database
        if levels is None:
            _, upper_ws = loading.get_quantile_range("h", 0.01, 0.99, 3)
            _, upper_hs = loading.get_quantile_range("hs", 0.01, 0.99, 3)
            levels = np.arange(
                0, upper_ws + 4 * upper_hs + 0.5 * self.step_size, self.step_size
            )

        # Calculate the boundaries
        h_boundaries = ProbabilityFunctions.calculate_boundaries(levels)

        # P(H=h, U=u, R=r | Q=q)
        p_hur_tr = model.calculate_probability_loading(
            result_variable="h",
            levels=h_boundaries,
            split_input_variables=["u"] + slow_stochastics + ["r"],
            model_uncertainty=self.model_uncertainty,
            given=slow_stochastics,
        )

        # Init wave overtopping statistics and loading
        wave_overtopping = WaveOvertopping(location, levels, p_hur_tr[1:-1])
        wave_overtopping_statistics = wave_overtopping.get_statistics()
        wave_overtopping_loading = wave_overtopping.get_loading()

        # Create an empty array for the levels and slow stochastics
        p_hbn_slow = np.zeros(
            (len(levels),)
            + tuple(
                [
                    getattr(wave_overtopping_statistics, f"n{x}")
                    for x in slow_stochastics
                ]
            )
        )

        # Iterate over the wave conditions model uncertainty
        # ASSUME: model uncertainties for wave height/period do not differ given the state of the barrier
        iterator = np.array(
            list(
                statistics.model_uncertainties.iterate_model_uncertainty_wave_conditions(
                    wave_period="tspec"
                )
            )
        )
        iterator = iterator if self.model_uncertainty else np.array([[1.0, 1.0, 1.0]])
        for n_id, (factor_hs, factor_tspec, p_occ) in enumerate(iterator):
            # Info
            print(
                f"[{datetime.now().strftime('%H:%M:%S')}]: Model uncertainties {n_id + 1}/{len(list(iterator))} (fhs = {round(factor_hs, 3)}; ftspec = {round(factor_tspec, 3)}; p = {round(p_occ, 3)})"
            )

            # Calculate the height of the HBNs
            wave_overtopping_loading.calculate_hbn(
                profile, self.q_overtopping, factor_hs, factor_tspec
            )

            # Repair HBN
            wave_overtopping_loading.repair_loadingmodels("hbn")

            # Calculate the probability of a HBN
            p_hbn_monz = wave_overtopping.calculate_probability_loading(
                result_variable="hbn",
                levels=levels[1:],
                model_uncertainty=False,
                split_input_variables=slow_stochastics,
                given=slow_stochastics,
            )
            p_hbn_slow += p_hbn_monz * p_occ

        # Translate to exceedance probabilities P(HBN>hbn | Q=q)
        ep_hbn_slow = np.cumsum(p_hbn_slow[::-1], axis=0)[-1::-1]

        # Process slow stochastics (they are always at the last axes of the matrix)
        if len(statistics.stochastics_slow) > 0:
            p_trapezoidal = model.process_slow_stochastics(ep_hbn_slow)
            exceedance_probability = p_trapezoidal * settings.periods_base_duration

        # If there are no slow stochastics, return the excedeence probability times the amount of periods
        else:
            exceedance_probability = ep_hbn_slow * settings.periods_block_duration

        # Return the exceedance frequency line of the HBN
        return FrequencyLine(levels, exceedance_probability)

    def set_critical_overtopping(self, q_overtopping: float):
        """
        Change the critical overtopping

        Parameters
        ----------
        q_overtopping : float
            The critical overtopping [m3/m/s]
        """
        # Cannot be smaller or equal to 0
        if q_overtopping <= 0:
            raise ValueError("[ERROR] Critical overtopping should be larger than 0.")

        # Save the critical overtopping
        self.q_overtopping = q_overtopping

    def set_levels(self, levels: list = None):
        """
        Change the levels.
        If levels is not defined, the frequency line is calculated based upon the 1st and 99th percentile.

        Parameters
        ----------
        levels : list, optional
            The levels at which the exceedance probability has to be calculated
        """
        self.levels = levels

    def set_step_size(self, step_size: float):
        """
        Change the step size of the frequency line.

        Parameters
        ----------
        step_size : float
            The step size of the frequency line
        """
        # Cannot be smaller or equal to 0
        if step_size <= 0:
            raise ValueError("[ERROR] Step size should be larger than 0.")

        # Save step size
        self.step_size = step_size

    def use_model_uncertainty(self, model_uncertainty: bool):
        """
        Use model uncertainty when calculating a frequency line.

        Parameters
        ----------
        model_uncertainty : bool
            Enable or disable the use of model uncertainties
        """
        self.model_uncertainty = model_uncertainty



import numpy as np

from ..statistics import Statistics
from ..stochastics.discrete_probability import DiscreteProbability
from ..stochastics.lake_level import LakeLevel
from ..stochastics.model_uncertainty import ModelUncertainty
from ..stochastics.wind_speed import WindSpeed
from ....settings.settings import Settings
from .....common.probability import ProbabilityFunctions


class StatisticsLake(Statistics):
    """
    Statistics class for the Lakes
    Water systems: IJssel Lake, Marker Lake, Veluwe Lakes and Grevelingen
    """

    def __init__(self, settings: Settings):
        """
        Init the Statistics class

        Parameters
        ----------
        settings : Settings
            The Settings object
        """
        # Inherit initialisation method from parent
        super().__init__(settings)

        # Wind
        self.wind_direction = DiscreteProbability(settings.wind_direction_probability)
        self.wind_speed = WindSpeed(settings)

        # Lake level
        self.lake_level = LakeLevel(settings)

        # Model uncertainty
        self.model_uncertainties = ModelUncertainty(settings)

        # Discrete, slow, fast stochatics
        self.stochastics_discrete = {
            "r": self.wind_direction.get_discretisation(),
            "k": [1],
        }
        self.stochastics_fast = {"u": self.wind_speed.get_discretisation()}
        self.stochastics_slow = {"a": self.lake_level.get_discretisation()}

    def calculate_probability(
        self, wind_direction: float, closing_situation: int = 1, given: list = []
    ):
        """
        Calculate the probability of occurence for the discretisation given the wind direction.

        Parameters
        ----------
        direction : float
            Wind direction
        closing_situation : int
            Closing situation, (irrelevant for Coast)
        given : list
            Given stochasts
        """
        # Wind speed
        ir = self.wind_direction.get_discretisation().tolist().index(wind_direction)
        p_wind = ProbabilityFunctions.probability_density(
            self.wind_speed.get_discretisation(),
            self.wind_speed.get_exceedance_probability()[:, ir],
        ).probability[:, None]
        if "u" in given:
            p_wind[:] = 1.0

        # Lake level
        if "a" in given:
            p_lake_level = np.ones_like(self.lake_level.get_discretisation())
        else:
            # If not given, use the instantaneous probability
            p_lake_level = ProbabilityFunctions.probability_density(
                self.lake_level.get_discretisation(),
                1 - self.lake_level.get_exceedance_probability(),
            ).probability

        # Probability of wind direction
        p_direction = 1.0 if "r" in given else self.wind_direction.get_probability()[ir]

        # Combine probabilities
        return p_wind * p_lake_level[None, :] * p_direction



from ..loading import Loading
from ..loading_model.loading_model import LoadingModel
from ....settings.settings import Settings
from .....io.database_hr import DatabaseHR


class LoadingLowerRiver(Loading):
    """
    Loading class for Lower Rivers
    Water systems: Tidal River Rhine, Tidal River Meuse
    """

    def __init__(self, settings: Settings):
        """
        Init the Loading object for the Coast

        Parameters
        ----------
        settings : Settings
            The Settings object
        """
        # Inherit the from parent
        super().__init__(settings)

        # Read and process the loading
        self.read_loading()

    def read_loading(self) -> None:
        """
        Read the HR result table and create LoadingModels
        """
        # Read table
        with DatabaseHR(self.settings.database_path) as database:
            table = database.get_result_table(self.settings)
            rvids = database.get_result_variables()
            ivids_west = database.get_input_variables()

        # For lower tidal rivers, there is a different loading model for sector east and west
        ivids_east = ivids_west.copy()
        ivids_east.remove("m")

        # Init LoadingModels for each combination of wind direction (r) and closing situation (k)
        for comb, deeltabel in table.groupby(["r", "k"]):
            wind_direction, closing_situation = comb

            # Sector west
            if ((wind_direction >= 0.0) and (wind_direction <= 11.25)) or (
                (wind_direction > 212.75) and (wind_direction <= 360.0)
            ):
                model = LoadingModel(
                    wind_direction, closing_situation, ivids_west, rvids
                )

            # Sector east
            else:
                model = LoadingModel(
                    wind_direction, closing_situation, ivids_east, rvids
                )

            # Init loading model
            model.initialise(deeltabel.copy())

            # Add model to the models dictionary
            self.model[comb] = model

        # Extend and repair loadingmodels
        self._extend_loadingmodels()
        self.repair_loadingmodels(rvids)



from .enum import WaterSystem


class CommonFunctions:
    """
    A class with common functions.
    """

    @staticmethod
    def is_coast(watersystem: WaterSystem) -> bool:
        """
        Returns True if a watersystem belongs to the coast loading model.

        Parameters
        ----------
        watersystem : WaterSystem
            The watersystem

        Returns
        -------
        bool
            True if the watersystem is Coast
        """
        if watersystem in [
            WaterSystem.COAST_SOUTH,
            WaterSystem.COAST_CENTRAL,
            WaterSystem.COAST_NORTH,
            WaterSystem.COAST_DUNES,
            WaterSystem.WADDEN_SEA_WEST,
            WaterSystem.WADDEN_SEA_EAST,
            WaterSystem.WESTERN_SCHELDT,
        ]:
            return True
        return False

    @staticmethod
    def is_lower_rivier(watersystem: WaterSystem) -> bool:
        """
        Returns True if a watersystem belongs to the lower river loading model.

        Parameters
        ----------
        watersystem : WaterSystem
            The watersystem

        Returns
        -------
        bool
            True if the watersystem is a lower river
        """
        if watersystem in [
            WaterSystem.MEUSE_TIDAL,
            WaterSystem.RHINE_TIDAL,
            WaterSystem.EUROPOORT,
        ]:
            return True
        return False

    @staticmethod
    def position_from_line(where, pt, refpts):
        """
        Check where a point is located relative to a line
        """

        def left(pt, refpts):
            A, B = refpts
            # Determine if the point is located left or right of the line
            sign = ((B[0] - A[0]) * (pt[1] - A[1])) - ((B[1] - A[1]) * (pt[0] - A[0]))
            if sign > 0:
                left = True
            else:
                left = False
            return left

        if where == "north" or where == "south":
            # Sorteer punten links naar recht, en kijk het links ligt
            if refpts[0][0] > refpts[1][0]:
                leftpoint, rightpoint = refpts[1], refpts[0]
            else:
                leftpoint, rightpoint = refpts[0], refpts[1]

            # Bekijk of het punt links of recht van de lijn links naar rechts ligt
            if where == "north":
                return left(pt, [leftpoint, rightpoint])
            else:
                return left(pt, [rightpoint, leftpoint])

        if where == "west" or where == "east":
            # Sorteer punten onder naar boven, en kijk het links ligt
            if refpts[0][1] > refpts[1][1]:
                downpoint, toppoint = refpts[1], refpts[0]
            else:
                downpoint, toppoint = refpts[0], refpts[1]

            if where == "west":
                return left(pt, [downpoint, toppoint])
            else:
                return left(pt, [toppoint, downpoint])



import numpy as np
import os

from ctypes import CDLL, POINTER, c_char_p, c_double, c_int, c_long, byref
from numpy.ctypeslib import ndpointer
from typing import Tuple


class Foreland:
    """
    This module will use the Dam and Foreland module (DaF) to transform wave conditions
    based on the schematized foreshore. The DaF module can be used to transform wave conditions
    over a breakwater and/or foreshore.
    """

    def __init__(self, profile, log: bool = False):
        # Path to the library
        lib_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "lib"))

        # Load the DaF library v20.1.1.692 (differ from Hydra-NL, there are some slight changes)
        self.daf_library = CDLL(os.path.join(lib_path, "DynamicLib-DaF.dll"))

        # Default settings
        self.alpha_c = c_double(1.0)
        self.fc_c = c_double(0.021)
        self.ratiodepth_c = c_double(0.5)
        self.minstepsize_c = c_double(1.0)
        self.invalid_c = c_double(-999.99)
        self.logging_c = c_int(int(log))
        self.loggingfilename_c = c_char_p("dlldaf_log.txt".encode("utf-8"))
        self.loggingfilenamelength_c = c_int(len(self.loggingfilename_c.value))
        self.g_c = c_double(9.81)
        self.rho_c = c_double(1000.0)

        # Intialize dll entries
        self.__initialize_dll_entries()

        # Set profile
        self.profile = profile

    def add_profile(self, profile) -> None:
        """
        Set profile

        Parameters
        ----------
        Profile : profile
            A Profile object
        """
        self.profile = profile

    def transform_wave_conditions(
        self,
        water_level: np.ndarray,
        significant_wave_height: np.ndarray,
        peak_wave_period: np.ndarray,
        wave_direction: np.ndarray,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Transform the wave conditions for the schematized foreland

        Parameters
        ----------
        water_level : np.ndarray
            Water level
        significant_wave_height : np.ndarray
            Significant wave height
        peak_wave_period : np.ndarray
            Peak wave period
        wave_direction : np.ndarray
            Wave direction

        Returns
        -------
        Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
            Water level and transformed wave conditions (h, hs, tp, dir)
        """
        # Get size and shape
        shp = significant_wave_height.shape

        # Determine part where wave height is larger than zero
        mask = (significant_wave_height > 0.0) & (peak_wave_period > 0.0)
        N = mask.sum()

        # Allocate output arrays as single dimension
        hm0dike = np.zeros(N, order="F")
        tpdike = np.zeros(N, order="F")
        refractedwaveangledike = np.zeros(N, order="F")
        message = "".ljust(1000).encode("utf-8")
        messagelength = c_int(1000)
        n_vl = (
            len(self.profile.foreland_x_coordinates)
            if self.profile.foreland_x_coordinates is not None
            else 1
        )
        x_vl = (
            np.array(self.profile.foreland_x_coordinates).astype(np.float64)
            if self.profile.foreland_x_coordinates is not None
            else np.array([0]).astype(np.float64)
        )
        y_vl = (
            np.array(self.profile.foreland_y_coordinates).astype(np.float64)
            if self.profile.foreland_x_coordinates is not None
            else np.array([-999]).astype(np.float64)
        )

        res = self.daf_library.C_FORTRANENTRY_RollerModel5(
            byref(c_int(self.profile.breakwater_type.value)),
            byref(c_double(self.profile.breakwater_level)),
            byref(self.alpha_c),
            byref(self.fc_c),
            byref(self.invalid_c),
            byref(c_int(N)),
            significant_wave_height[mask].astype(np.float64),
            peak_wave_period[mask].astype(np.float64),
            water_level[mask].astype(np.float64),
            wave_direction[mask].astype(np.float64),
            byref(c_double(self.profile.dike_orientation)),
            byref(c_int(n_vl)),
            x_vl,
            byref(self.minstepsize_c),
            y_vl,
            byref(self.ratiodepth_c),
            byref(self.logging_c),
            self.loggingfilename_c,
            self.loggingfilenamelength_c,
            byref(self.g_c),
            byref(self.rho_c),
            hm0dike,
            tpdike,
            refractedwaveangledike,
            message,
            messagelength,
        )
        message = message.decode("utf-8")
        message = message.rstrip()
        messagelength = len(message)

        if res != 0:
            print(message + " - Using uncorrected wave parameters.")
            print(self.profile.foreland_x_coordinates)
            print(self.profile.foreland_y_coordinates)
            hm0dike[:] = significant_wave_height[mask].ravel()[:]
            tpdike[:] = peak_wave_period[mask].ravel()[:]
            refractedwaveangledike[:] = wave_direction[mask].ravel()[:]

        # If not all input conditions were non-zero, put the calculated conditions on the original grid again.
        if not mask.all():
            # Copy original values
            hm0_tmp, tp_tmp, wdir_tmp = (
                significant_wave_height.copy(),
                peak_wave_period.copy(),
                wave_direction.copy(),
            )

            # Insert calculated values
            hm0_tmp[mask] = hm0dike
            tp_tmp[mask] = tpdike
            wdir_tmp[mask] = refractedwaveangledike
            conditions = (water_level, hm0_tmp, tp_tmp, wdir_tmp % 360.0)

        # Else, all values, where calculated. Only reshape to input shape
        else:
            conditions = (
                water_level,
                hm0dike.reshape(shp),
                tpdike.reshape(shp),
                refractedwaveangledike.reshape(shp) % 360.0,
            )

        # Return the transformed conditions
        return conditions

    def __initialize_dll_entries(self) -> None:
        """
        Initializes the arguments types of various functions in the dynamic library.
        """
        # get entry point in dll, the function to use
        arraypointer = ndpointer(dtype="double", ndim=1, flags="F_CONTIGUOUS")

        # Define all the argument types
        argtypes = {
            "DamType": POINTER(c_int),
            "DamHeight": POINTER(c_double),
            "Alpha": POINTER(c_double),
            "Fc": POINTER(c_double),
            "Invalid": POINTER(c_double),
            "DimHm0": POINTER(c_int),
            "Hm0": arraypointer,
            "Tp": arraypointer,
            "Wlev": arraypointer,
            "IncomingWaveAngle": arraypointer,
            "DikeNormal": POINTER(c_double),
            "DimX": POINTER(c_int),
            "X": arraypointer,
            "MinStepSize": POINTER(c_double),
            "BottomLevel": arraypointer,
            "RatioDepth": POINTER(c_double),
            "Logging": POINTER(c_int),
            "LoggingFileName": c_char_p,
            "LoggingFileNameLength": c_int,
            "Ag": POINTER(c_double),
            "Rho": POINTER(c_double),
            "Hm0Dike": arraypointer,
            "TpDike": arraypointer,
            "RefractedWaveAngleDike": arraypointer,
            "Message": c_char_p,
            "MessageLength": c_int,
        }

        # Note function definition for DAF module ROLLERMODEL5
        self.daf_library.C_FORTRANENTRY_RollerModel5.restype = c_long
        self.daf_library.C_FORTRANENTRY_RollerModel5.argtypes = [
            argtypes[name]
            for name in [
                "DamType",
                "DamHeight",
                "Alpha",
                "Fc",
                "Invalid",
                "DimHm0",
                "Hm0",
                "Tp",
                "Wlev",
                "IncomingWaveAngle",
                "DikeNormal",
                "DimX",
                "X",
                "MinStepSize",
                "BottomLevel",
                "RatioDepth",
                "Logging",
                "LoggingFileName",
                "LoggingFileNameLength",
                "Ag",
                "Rho",
                "Hm0Dike",
                "TpDike",
                "RefractedWaveAngleDike",
                "Message",
                "MessageLength",
            ]
        ]

        # Note function definition for DAF function to obtain version number.
        self.daf_library.GetVersionInfo.argtypes = [c_char_p, c_int]



from abc import ABC, abstractmethod
from typing import Union

from ..hrdatabase.hrdatabase import HRDatabase
from ..location.location import Location


class Calculation(ABC):
    """
    Base class for calculation modules
    """

    @abstractmethod
    def __init__(self):
        """
        Init class, differs per calculation
        """
        pass

    def calculate(self, input: Union[Location, HRDatabase]):
        """
        Execute a calculation

        Parameters
        ----------
        input : Union[Location, HRDatabase]
            The input
        """
        # Depending on the type of input, call CalculateLocation
        if isinstance(input, Location):
            return self.calculate_location(input)

        elif isinstance(input, HRDatabase):
            return {
                loc: self.calculate_location(input.get_location(loc)) for loc in input
            }

        else:
            raise NotImplementedError("[ERROR] Input type not implemented")

    @abstractmethod
    def calculate_location(self, location: Location):
        """
        Executes a calculation for a location

        Parameter
        ---------
        location : Location
            The Location object
        """
        pass



import numpy as np

from .sea_level import SeaLevel
from .....settings.settings import Settings
from ......common.interpolate import Interpolate
from ......common.probability import ProbabilityFunctions
from ......io.file_hydranl import FileHydraNL


class SeaLevelLowerRiver(SeaLevel):
    """
    Class to describe the sea (water) level statistics.
    Sea level statistics are conditional on the wind direction.
    """

    def __init__(self, settings: Settings):
        """
        Initialise the sea level statistics for Rhine Tidal, Meuse Tidal,
        EuroPoort, Volkerak-Zoommeer and Hollandsche IJssel.

        Parameters
        ----------
        settings : Settings
            The Settings object
        """
        # Inherit
        super().__init__()

        # Read the exceedance probability of the sea level conditional to the wind direction
        m, epm = FileHydraNL.read_file_ncolumns(settings.sea_level_probability)
        m += settings.sea_level_rise
        nr = epm.shape[1]

        # Correct for the distance between Hoek van Holland and Maasmond
        self.translation_m = 0.02 - settings.sea_level_rise

        # Equidistant filling vector with seawater levels
        self.m = ProbabilityFunctions.get_hnl_disc_array(
            settings.m_min, settings.m_max, settings.m_step
        )

        # Alloceer matrices
        self.nm = len(self.m)
        self.epm = np.zeros([self.nm, nr])

        # Interpolate the conditional exceedance probability of the sea level onto the given grid
        for ir in range(nr):
            iszero = epm[:, ir] == 0.0
            if iszero.any():
                zeroidx = np.where(iszero)[0][-1]

                # Interpoleer of extrapoleer logaritmisch tot de 0
                self.epm[:zeroidx, ir] = np.exp(
                    Interpolate.inextrp1d(
                        x=self.m[:zeroidx], xp=m[~iszero], fp=np.log(epm[~iszero, ir])
                    )
                )

                # Interpoleer of extrapoleer lineair vanaf 0
                self.epm[zeroidx:, ir] = np.maximum(
                    0.0, Interpolate.inextrp1d(self.m[zeroidx:], m, epm[:, ir])
                )

            else:
                self.epm[:, ir] = np.exp(
                    Interpolate.inextrp1d(x=self.m, xp=m, fp=np.log(epm[:, ir]))
                )

        # Limit the upper limit of the excedaance probability to 1.0
        self.epm = np.minimum(1.0, self.epm)

        # Scaling of the conditional exceedance probability to the used time period
        if (settings.block_duration != 12.0) and (settings.block_duration is not None):
            self.epm = 1.0 - (1.0 - self.epm) ** (settings.block_duration / 12.0)

        #  Allocate matrix
        self.pm = np.zeros((self.nm, nr))

        #  Berekenen kansdichtheid zeewaterstand gegeven de windrichting
        for ir in range(nr):
            self.pm[:, ir] = ProbabilityFunctions.probability_density(
                self.m, self.epm[:, ir]
            ).density

        #  Inlezen parameters kansverdeling gezamenlijke kansdichtheid zeewaterstand, windsnelheid
        #  en windrichting
        _, self.pwinds = FileHydraNL.read_file_ncolumns(
            settings.wind_sea_level_probability
        )
        self.pwinds = self.pwinds.T

        #  Geef twee instellingen van het gegevensblok mee in de structure self
        self.transitional_wind = settings.transitional_wind



import numpy as np

from .wave_shape import WaveShape
from ....settings.settings import Settings
from .....common.enum import WaveShapeType
from .....common.interpolate import Interpolate
from .....io.file_hydranl import FileHydraNL


class Discharge:
    """
    Class to describe the discharge statistics.
    """

    def __init__(self, settings: Settings):
        """
        Constructor for discharge statistics

        Parameters
        ----------
        settings : Settings
            The Settings object
        """
        # Read exceedance probability peak discharge
        qpeak, epqeak = FileHydraNL.read_file_2columns(settings.discharge_probability)

        # Create grid for peak discharges
        self.qpeak = np.r_[
            np.arange(settings.q_min, settings.q_max, settings.q_step), settings.q_max
        ]
        self.nqpeak = len(self.qpeak)

        # Inter/extrapolate the exceedance probability of the discharge
        self.epqpeak = np.exp(
            Interpolate.inextrp1d(x=self.qpeak, xp=qpeak, fp=np.log(epqeak))
        )

        #  Transformatie van de tabel met OVERschrijdingskansen naar exponentiële ruimte
        #  Zie ook vgl. 4.5 van [Geerse, 2003], merk op dat het daar ONDERschrijdingskansen betreft#
        self.eqpeak_exp = -np.log(self.epqpeak)

        # Init the wave forms
        self.wave_shape = WaveShape(settings, type=WaveShapeType.DISCHARGE)
        self.wave_shape.initialise_wave_shapes(self.qpeak)

        # Create grid with blok discharges
        if settings.q_min < np.min(self.qpeak):
            step = self.qpeak[1] - self.qpeak[0]
            self.qblok = np.r_[
                np.arange(settings.q_min, np.min(self.qpeak), step), self.qpiek
            ]
        elif settings.q_min > np.min(self.qpeak):
            raise ValueError("[ERROR] Q_min is larger than the lowest Q_peak.")
        else:
            self.qblok = self.qpeak
        self.nqblok = len(self.qblok)

        # Calculate the instantaneous exceedance probability
        inst_epq = self.wave_shape.instantaneous_exceedance_probability(
            self.epqpeak, self.qblok
        )

        # Make sure the exceedance probability is not larger than 1 or smaller than 0
        self.upqblok = np.clip(1.0 - inst_epq, 0.0, 1.0)

    def __len__(self):
        """
        Return the number of discharge discretisations.

        Returns
        -------
        int
            Number of discretisations
        """
        return self.nqblok

    def get_discretisation(self) -> np.ndarray:
        """
        Return the discharge discretisation.

        Returns
        -------
        np.ndarray
            1D array with discretisation
        """
        return self.qblok

    def get_exceedance_probability(self) -> np.ndarray:
        """
        Return exceedance probility of the discharge

        Returns
        -------
        np.ndarray
            A 1D array with the discharge exceedance probability
        """
        return self.upqblok

    def get_wave_shape(self) -> WaveShape:
        """
        Return the wave form statistics object.

        Returns
        -------
        WaveShape
            The WaveShape statistics
        """
        return self.wave_shape



import numpy as np

from ..statistics import Statistics
from ..stochastics.barrier.barrier_ramspol import BarrierRamspol
from ..stochastics.discrete_probability import DiscreteProbability
from ..stochastics.discharge import Discharge
from ..stochastics.lake_level import LakeLevel
from ..stochastics.model_uncertainty import ModelUncertainty
from ..stochastics.wind_speed import WindSpeed
from ....settings.settings import Settings
from .....common.probability import ProbabilityFunctions


class StatisticsIJsselVechtdelta(Statistics):
    """
    Statistics class for the IJssel-Vechtdelta
    Water systems: IJssel Delta, Vecht Delta
    """

    def __init__(self, settings: Settings):
        """
        Init the Statistics class

        Parameters
        ----------
        settings : Settings
            The Settings object
        """
        # Inherit initialisation method from parent
        super().__init__(settings)

        # Wind
        self.wind_direction = DiscreteProbability(settings.wind_direction_probability)
        self.wind_speed = WindSpeed(settings)

        # Lake level
        self.lake_level = LakeLevel(settings)

        # Discharge
        self.discharge = Discharge(settings)

        # Combined probability(density) of lake level and discharge
        self.density_aq_peak, self.density_aq, self.ovduration_aq = (
            self.__prob_combined()
        )

        # Ramspol
        self.barrier = BarrierRamspol(settings, self.wind_direction)

        # Model uncertainty
        self.model_uncertainties = ModelUncertainty(settings)

        # Discrete, slow, fast stochatics
        self.stochastics_discrete = {
            "r": self.wind_direction.get_discretisation(),
            "k": self.barrier.k,
        }
        self.stochastics_fast = {"u": self.wind_speed.get_discretisation()}
        self.stochastics_slow = {
            "a": self.lake_level.get_discretisation(),
            "q": self.discharge.get_discretisation(),
        }

    def calculate_probability(
        self, wind_direction: float, closing_situation: int = 1, given: list = []
    ):
        """
        Calculate the probability of occurence for the discretisation given the wind direction.

        Parameters
        ----------
        direction : float
            Wind direction
        closing_situation : int
            Closing situation, (irrelevant for Coast)
        given : list
            Given stochasts
        """
        # Probility wind speed
        ir = self.wind_direction.get_discretisation().tolist().index(wind_direction)
        kanswind = ProbabilityFunctions.probability_density(
            self.wind_speed.get_discretisation(),
            self.wind_speed.get_exceedance_probability()[:, ir],
        ).probability
        if "u" in given:
            kanswind[:] = 1.0

        # Lake level and discharge
        if ("a" in given) and ("q" in given):
            # If both are conditional, fill the matrix with ones
            kans_aq = np.ones((len(self.lake_level), len(self.discharge)))
        else:
            # If one of two is not conditional, use the momentary probability
            kans_aq = self.density_aq
            if "a" not in given:
                kans_aq[...] = ProbabilityFunctions.conditional_probability(
                    kans_aq, axis=0
                )
            if "q" not in given:
                kans_aq[...] = ProbabilityFunctions.conditional_probability(
                    kans_aq, axis=1
                )

        # Probility for wind direction
        kanswr = 1.0 if "r" in given else self.wind_direction.get_probability()[ir]

        # Combine all probabilities
        kansen = kanswind[:, None, None] * kans_aq[None, :, :] * kanswr

        # Ramspol closing probability
        keringkans = self.barrier.calculate_closing_probability(None, closing_situation)
        kansen *= keringkans

        # Swap axis
        kansen = np.swapaxes(kansen, 1, 2)

        # Return probability
        return kansen

    def __prob_combined(self) -> np.ndarray:
        """Bereken de gezamelijke kans op afvoer en meerpeil. De functie wordt zowel
        gebruikt in de VIJD als het VZM, vandaar dat deze niet aan een klasse gekoppeld is.

        Parameters
        ----------
        afvoer : afvoer.Afvoer
            Afvoerstatistiekobject
        meerpeil : meerpeil.Meerpeil
            Meerpeilstatistiekobject
        invoergeg:
            Invoergegevensobject

        Returns
        -------
        np.ndarray
            Gezamelijke kansdichtheid van meerpeil en afvoer
        """
        # Bereken de kansdichtheid van de piekafvoer
        kansafvoer = ProbabilityFunctions.probability_density(
            self.discharge.qblok, self.discharge.epqpeak
        )
        kansmeerpeil = ProbabilityFunctions.probability_density(
            self.lake_level.ablok, self.lake_level.k_apeak
        )

        ovmaxm = self.lake_level.epapeak[-1]
        sigma = self.settings.sigma_aq

        # Bereken de gezamenlijke kansdichtheid van de piekwaardes van de afvoer en het meerpeil
        assert self.lake_level.nablok == self.lake_level.napeak
        nm = self.lake_level.nablok
        assert self.discharge.nqblok == self.discharge.nqpeak
        nq = self.discharge.nqblok
        dichtheid_mqpiek = np.full((nm, nq), 0.0)

        for iq in range(nq):
            somkans = 0.0

            for im in reversed(range(nm)):
                deltam = kansmeerpeil.delta[im]

                y = self.lake_level.k_apeak[im] - self.discharge.eqpeak_exp[iq]

                dichtheid_mqpiek[im, iq] = (
                    kansmeerpeil.density[im]
                    * np.exp(-(y**2) / (2.0 * sigma**2))
                    / (sigma * (2.0 * np.pi) ** 0.5)
                )

                # Als de som over de kansen groter wordt dan 1, dan wordt voor het beschouwde, die kansmassa
                # aangehouden dat de som over de kansen gelijk wordt aan 1. De verdere blokjes krijgen geen kans.
                if (somkans <= (1.0 - ovmaxm)) & (
                    (somkans + dichtheid_mqpiek[im, iq] * deltam) >= (1.0 - ovmaxm)
                ):
                    dichtheid_mqpiek[im, iq] = (1.0 - ovmaxm - somkans) / deltam
                    somkans = 1.0 - ovmaxm
                    break

                somkans += dichtheid_mqpiek[im, iq] * deltam

            if somkans > (1.0 - ovmaxm):
                dichtheid_mqpiek[:, iq] /= somkans

            dichtheid_mqpiek[:, iq] *= kansafvoer.density[iq]

        # En de blokstat
        # Overschrijdingsduren van afvoer- en meerpeilgolf bepalen
        overschrijdingsduren_mq = (
            self.discharge.wave_shape.bepaal_gezamenlijke_overschrijding(
                golfvormen_st1=self.lake_level.wave_shape,
                niveaus_st1=self.lake_level.ablok,
                golfvormen_st2=self.discharge.wave_shape,
                niveaus_st2=self.discharge.qblok,
            )
        )

        # Momentane kansen bepalen
        dq = ProbabilityFunctions.probability_density(
            self.discharge.qpeak, self.discharge.epqpeak
        ).delta
        dm = ProbabilityFunctions.probability_density(
            self.lake_level.apeak, self.lake_level.epapeak
        ).delta
        kans_mqpiek = dichtheid_mqpiek * dm[:, None] * dq[None, :]
        ovkansmq = (kans_mqpiek[:, :, None, None] * overschrijdingsduren_mq).sum(
            (0, 1)
        ) / (self.settings.base_duration)

        # Corrigeer kleine onvolkomenheden
        ovkansmq = np.maximum.accumulate(ovkansmq[::-1])[::-1]

        tussenkans = np.zeros((nq, nm))
        for iq in range(nq):
            tussenkans[iq, :] = ProbabilityFunctions.probability_density(
                self.lake_level.ablok, ovkansmq[:, iq]
            ).probability

        # Corrigeer kleine onvolkomenheden
        tussenkans = np.maximum.accumulate(tussenkans[::-1])[::-1]

        dichtheid_mqblok = np.zeros((nm, nq))
        for im in range(nm):
            dichtheid_mqblok[im, :] = ProbabilityFunctions.probability_density(
                self.discharge.qblok, tussenkans[:, im]
            ).probability

        dichtheid_mqblok /= dichtheid_mqblok.sum()

        return dichtheid_mqpiek, dichtheid_mqblok, overschrijdingsduren_mq



import os
import pandas as pd
import sqlite3

from typing import List
from typing import Union

from .database_settings import DatabaseSettings
from ..common.enum import WaterSystem
from ..location.settings.settings import Settings


class DatabaseHR:
    """
    HR database sqlite
    """

    def __init__(self, path_to_database: str) -> None:
        # Check if the path is valid
        if not os.path.exists(path_to_database):
            raise OSError(path_to_database)

        # Save the path
        self.path_to_database = path_to_database
        self.con = None

    def __enter__(self) -> "DatabaseHR":
        # Init the connection
        self.con = sqlite3.connect(self.path_to_database)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        # Close the connection
        self.con.close()

    def get_water_system(self) -> WaterSystem:
        """
        Obtain the water system from the .sqlite database

        Returns
        -------
        WaterSystem
            Corresponding water system
        """
        # Obtain the water system ID from the sqlite
        wsid = self.con.execute("SELECT GeneralId FROM General").fetchone()[0]

        # Return the WaterSystem
        return WaterSystem(wsid)

    def get_hrdlocations_names(self) -> List[str]:
        """
        Obtain a list with all names of the hrdlocations

        Returns
        -------
        list[str]
            A list with all names of hrdlocations
        """
        # Obtain the water system ID from the sqlite
        hrdlocations = self.con.execute("SELECT Name FROM HRDLocations").fetchall()

        # Convert the result to a list of strings
        names = [row[0] for row in hrdlocations]

        # Return names
        return names

    def get_hrdlocation_id(self, hrdlocation: Union[str, Settings]) -> int:
        """
        Returns the HRDLocationID

        Parameters
        ----------
        hrdlocation : Union[str, Settings]
            HRDLocation

        Returns
        -------
        int
            HRDLocationId
        """
        # Obtain the HRDLocationName from a Settings object
        if isinstance(hrdlocation, Settings):
            hrdlocation = hrdlocation.location

        # Obtain the water system ID from the sqlite
        hrdlocationid = self.con.execute(
            f"SELECT HRDLocationId FROM HRDLocations WHERE Name = '{hrdlocation}'"
        ).fetchone()[0]

        # Return HRDLocationId
        return hrdlocationid

    def get_hrdlocation_xy(
        self, hrdlocation: Union[str, Settings]
    ) -> Union[float, float]:
        """
        Returns the X and Y coordinate of the HRDLocation

        Parameters
        ----------
        hrdlocation : Union[str, Settings]
            HRDLocation

        Returns
        -------
        Union[float, float]
            X and Y coordinate
        """
        # Obtain the HRDLocationName from a Settings object
        if isinstance(hrdlocation, Settings):
            hrdlocation = hrdlocation.location

        # Obtain the water system ID from the sqlite
        hrdlocationid = self.con.execute(
            f"SELECT XCoordinate, YCoordinate FROM HRDLocations WHERE Name = '{hrdlocation}'"
        ).fetchone()

        # Return HRDLocationId
        return hrdlocationid

    def get_input_variables(self) -> list:
        """
        Return the input variables

        Returns
        -------
        list
            List with input variables
        """
        # Query
        sql = "SELECT InputVariableId FROM HRDInputVariables"
        data = self.con.execute(sql).fetchall()

        # Settings database
        with DatabaseSettings() as database:
            ivids = database.get_input_variable_ids()
        data = [ivids[i[0]] for i in data]

        # Shift wind speed in front
        if "u" in data:
            data.pop(data.index("u"))
            data.insert(0, "u")

        # Return results
        return data

    def get_result_variables(self) -> list:
        """
        Return the result variables

        Returns
        -------
        list
            List with result variables
        """
        # Query
        sql = "SELECT ResultVariableId FROM HRDResultVariables"
        data = self.con.execute(sql).fetchall()

        # Settings database
        with DatabaseSettings() as database:
            rvids = database.get_result_variable_ids()
        data = [rvids[i[0]] for i in data]

        # For the coast, if not defined, the local water level is equal to the sea level
        if self.get_water_system() in [
            WaterSystem.WADDEN_SEA_EAST,
            WaterSystem.WADDEN_SEA_WEST,
            WaterSystem.COAST_NORTH,
            WaterSystem.COAST_CENTRAL,
            WaterSystem.COAST_SOUTH,
            WaterSystem.WESTERN_SCHELDT,
        ]:
            if "h" not in rvids:
                data.insert(0, "h")

        # Return results
        return data

    def get_model_uncertainties(
        self, hrdlocation: Union[int, str, Settings]
    ) -> pd.DataFrame:
        """
        Return the model uncertainties

        Parameters
        ----------
        hrdlocation : Union[int, str, Settings]
            HRDLocation in form of HRDLocationId, HRDLocationName or Settings object

        Returns
        -------
        pd.DataFrame
            DataFrame with the distribution per closing situation
        """
        # Obtain the hrdlocationid
        if isinstance(hrdlocation, (str, Settings)):
            hrdlocation = self.get_hrdlocation_id(hrdlocation)

        # Obtain all model uncertainties from the database for the hrdlocation
        sql = f"""
                SELECT umf.HRDLocationId, umf.ClosingSituationId, hrv.ResultVariableId, umf.Mean, umf.Standarddeviation
                FROM UncertaintyModelFactor umf
                INNER JOIN HRDResultVariables hrv
                ON umf.HRDResultColumnId = hrv.HRDResultColumnId
                WHERE umf.HRDLocationId = {hrdlocation}
                """
        data = pd.read_sql(sql, self.con, index_col="HRDLocationId")

        # Adjust dataframe
        with DatabaseSettings() as database:
            rvids = database.get_result_variable_ids()
        data.rename(
            columns={
                "ClosingSituationId": "k",
                "ResultVariableId": "rvid",
                "Mean": "mean",
                "Standarddeviation": "stdev",
            },
            inplace=True,
        )
        data["rvid"].replace(rvids, inplace=True)

        # Return the model uncertainties
        return data

    def get_correlation_uncertainties(
        self, hrdlocation: Union[int, str, Settings]
    ) -> pd.DataFrame:
        """
        Return the correlation between model uncertainties

        Parameters
        ----------
        hrdlocation : Union[int, str, Settings]
            HRDLocation in form of HRDLocationId, HRDLocationName or Settings object
        """
        # Obtain the hrdlocationid
        if isinstance(hrdlocation, (str, Settings)):
            hrdlocation = self.get_hrdlocation_id(hrdlocation)

        # ResultVariableIds
        with DatabaseSettings() as database:
            rvids = database.get_result_variable_ids()

        # Data uit correlatie tabel
        try:
            sql = f"""
                    SELECT ucf.HRDLocationId, ucf.ClosingSituationId, hrv.ResultVariableId, ucf.HRDResultColumnId2, ucf.Correlation
                    FROM UncertaintyCorrelationFactor ucf
                    INNER JOIN HRDResultVariables hrv
                    ON ucf.HRDResultColumnId = hrv.HRDResultColumnId
                    WHERE ucf.HRDLocationId = {hrdlocation}
                    """
            data = pd.read_sql(sql, self.con, index_col="HRDLocationId")

            # Vertaal tabel naar HRDResultColumnId2
            # Zo niet, negeer en ga verder, neem aan dat de HRDResultColumnId2 heeft dezelfde Ids als HRDResultColumnId
            try:
                sql = """
                        SELECT HRDResultColumnId2, ResultVariableId
                        FROM HRDResultVariables2 hrv2
                        INNER JOIN HRDResultVariables hrv ON hrv.HRDResultColumnId = hrv2.HRDResultColumnId
                        """
                data_hrdid2 = self.con.execute(sql).fetchall()
                hrdid2_to_rvid = {_hrdid: _hrdid2 for _hrdid, _hrdid2 in data_hrdid2}
                data = data.replace({"HRDResultColumnId2": hrdid2_to_rvid})
            except Exception as e:
                print(f"ERROR: {e}, continuing without")
                pass

        # Geen correlaties aanwezig, return leeg dataframe
        except Exception as e:
            print(f"ERROR: {e}, continuing without correlation")
            data = pd.DataFrame(
                columns=[
                    "HRDLocationId",
                    "ClosingSituationId",
                    "ResultVariableId",
                    "HRDResultColumnId2",
                    "Correlation",
                ]
            )

        # Replace column names
        data.rename(
            columns={
                "ClosingSituationId": "k",
                "ResultVariableId": "rvid",
                "HRDResultColumnId2": "rvid2",
                "Correlation": "rho",
            },
            inplace=True,
        )

        # Check of alle ResultVariableId(2) rvids zijn
        if not set(data["rvid"]).issubset(set(rvids)) or not set(
            data["rvid2"]
        ).issubset(set(rvids)):
            raise ValueError("ERROR")

        # Change ResultVariableId(2) to rvids
        data["rvid"].replace(rvids, inplace=True)
        data["rvid2"].replace(rvids, inplace=True)

        # Return the model uncertainties
        return data

    def get_wind_directions(self) -> dict:
        """
        Obtain a dictionary with HRDWindDirectionIds and Directions

        Returns
        -------
        dict
            A dictionary {HRDWindDirectionId : Direction}
        """
        # Wind directions
        results = self.con.execute("SELECT * FROM HRDWindDirections").fetchall()

        # Process wind directions such that {wind_id : wind_direction}
        wind_direction = {wid: wr for wid, wr in results}

        # Return wind direction dictionary
        return wind_direction

    def get_result_table(self, hrdlocation: Union[str, Settings]) -> pd.DataFrame:
        """
        Function to read the load combinations of a location to a pandas DataFrame

        Parameters
        ----------
        hrdlocation : Union[str, Settings]
            HRDLocation

        Returns
        -------
        pd.DataFrame
            A DataFrame with load combinations
        """
        # Obtain HRDLocationId
        hrdlocationid = self.get_hrdlocation_id(hrdlocation)
        with DatabaseSettings() as database:
            ivids = database.get_input_variable_ids()
            rvids = database.get_result_variable_ids()

        # Obtain all data from the HydroDynamicData table
        # (HydroDynamicDataId, HRDLocationId, ClosingSituationId, HRDWindDirectionID)
        query = f"SELECT * FROM HydroDynamicData WHERE HRDLocationId = {hrdlocationid}"
        hydrodynamicdata = pd.read_sql(query, self.con, index_col="HydroDynamicDataId")

        # Obtain all data from the HydroDynamicInputData table
        hydrodynamicdataids = ",".join(
            hydrodynamicdata.index.values.astype(str).tolist()
        )
        query = """
        SELECT ID.HydroDynamicDataId, IV.InputVariableId, ID.Value
        FROM HydroDynamicInputData ID
        INNER JOIN HRDInputVariables IV ON ID.HRDInputColumnId = IV.HRDInputColumnId
        WHERE HydroDynamicDataId IN ({});
        """.format(hydrodynamicdataids)
        hydrodynamicinputdata = pd.read_sql(
            query, self.con, index_col=["HydroDynamicDataId", "InputVariableId"]
        ).unstack()
        ivcols = [ivids[i] for i in hydrodynamicinputdata.columns.get_level_values(1)]
        hydrodynamicinputdata.columns = ivcols

        # Obtain all data from the HydroDynamicResultData table
        query = """
        SELECT RD.HydroDynamicDataId, RV.ResultVariableId, RD.Value
        FROM HydroDynamicResultData RD
        INNER JOIN HRDResultVariables RV ON RD.HRDResultColumnId = RV.HRDResultColumnId
        WHERE HydroDynamicDataId IN ({});
        """.format(hydrodynamicdataids)
        hydrodynamicresultdata = pd.read_sql(
            query, self.con, index_col=["HydroDynamicDataId", "ResultVariableId"]
        ).unstack()
        rvcols = [rvids[i] for i in hydrodynamicresultdata.columns.get_level_values(1)]
        hydrodynamicresultdata.columns = rvcols

        # Merge the three tables
        results = hydrodynamicdata.join(hydrodynamicinputdata).join(
            hydrodynamicresultdata
        )

        # Vervang windrichting
        windrdict = self.get_wind_directions()
        for wid, r in windrdict.items():
            if r == 0.0:
                windrdict[wid] = 360.0
        results["Wind direction"] = [
            windrdict[i] for i in results["HRDWindDirectionId"].array
        ]
        results.drop(["HRDLocationId", "HRDWindDirectionId"], axis=1, inplace=True)

        # Replace discrete stochasts
        results.rename(
            columns={"Wind direction": "r", "ClosingSituationId": "k"}, inplace=True
        )

        # Move r to the front
        results.insert(0, "r", results.pop("r"))
        results.sort_values(by=["k", "r"] + ivcols, inplace=True)

        # Return results
        return results

    def get_closing_levels_table_europoort(self) -> dict:
        """
        Read the closing levels for the Europoort barrier.

        Returns
        -------
        pd.DataFrame
            A Dataframe with the closing level at sea (m) given r, u, q
        """

        # If there is a table called 'Sluitfunctie Europoortkering', use it
        try:
            # Read the table
            table = pd.read_sql(
                "SELECT * FROM [Sluitfunctie Europoortkering]", con=self.con
            )

        # Otherwise use the default functions
        except Exception as e:
            print(f"{e}: Using default functions")
            PATH = os.path.join(
                os.path.split(os.path.dirname(__file__))[0],
                "data",
                "statistics",
                "Sluitpeilen",
            )
            if self.get_water_system() in [
                WaterSystem.RHINE_TIDAL,
                WaterSystem.EUROPOORT,
            ]:
                table = pd.read_csv(
                    os.path.join(PATH, "Sluitfunctie Europoortkering Rijn 2017.csv"),
                    delimiter=";",
                )
            elif self.get_water_system() == WaterSystem.MEUSE_TIDAL:
                table = pd.read_csv(
                    os.path.join(PATH, "Sluitfunctie Europoortkering Maas 2017.csv"),
                    delimiter=";",
                )
            else:
                raise (
                    f"[ERROR] No closing levels for water system '{self.get_water_system().name}'."
                )

        # All columns to lower
        table.columns = table.columns.str.lower()

        # Rename
        table.rename(
            columns={
                "windrichting": "r",
                "afvoer": "q",
                "windsnelheid": "u",
                "zeewaterstand": "m",
            },
            inplace=True,
        )

        # Return table
        return table

    def get_closing_levels_table_eastern_scheldt(self):
        """
        Read the closing levels for the Eastern Scheldt barrier.

        Returns
        -------
        pd.DataFrame
            A Dataframe with the closing level (h_rpb), given r, u, m, d, p
        """
        # Read the table from the ClosingCriterionsOSK
        table = pd.read_sql("SELECT * FROM ClosingCriterionsOSK", con=self.con)

        # Rename the entries
        table.rename(
            columns={
                "WindDirection": "r",
                "WindSpeed": "u",
                "WaterLevel": "m",
                "StormDuration": "d",
                "PhaseDifference": "p",
                "WaterLevelRPB": "h_rpb",
            },
            inplace=True,
        )

        # Return table
        return table

    def get_closing_situations_eastern_scheldt(self) -> dict:
        """
        Read the closing situations from the database (ClosingSituationId : (Description : FailingLocks)).
        e.g. 1 : ("Reguliere sluiting", 0)

        Only works for the Eastern Scheldt.
        """
        # Check watersystem
        if self.get_water_system() != WaterSystem.EASTERN_SCHELDT:
            raise ValueError(
                "[ERROR] Function can only be called for the Eastern Scheldt"
            )

        # Read table
        sql = """
                SELECT C.ClosingSituationId, T.Description, C.FailingLocks
                FROM ClosingSituations C
                INNER JOIN ClosingSituationTypes T ON C.ClosingSituationTypeId = T.ClosingSituationTypeId
                """
        results = self.con.execute(sql).fetchall()

        # Post processing into an dictionary
        results = {i[0]: (i[1], i[2]) for i in results}

        # Return
        return results

    def get_result_table_eastern_scheldt(
        self, hrdlocation: Union[str, Settings]
    ) -> pd.DataFrame:
        """
        Function to export the loadcombinations of a location to a pandas DataFrame

        Parameters
        ----------
        naam : str
            Locationname
        """
        # Obtain HRDLocationId
        hrdlocationid = self.get_hrdlocation_id(hrdlocation)
        with DatabaseSettings() as database:
            ivids = database.get_input_variable_ids()
            rvids = database.get_result_variable_ids()

        # First collect the dataids. Also replace wind direction ids with real ids
        SQL = """
        SELECT D.HydraulicLoadId, D.ClosingSituationId, W.Direction AS "Wind direction"
        FROM HydroDynamicData D
        INNER JOIN HRDWindDirections W ON D.HRDWindDirectionId=W.HRDWindDirectionId;"""
        dataids = pd.read_sql(SQL, self.con, index_col="HydraulicLoadId")
        dataids.rename(
            columns={"Wind direction": "r", "ClosingSituationId": "k"}, inplace=True
        )

        # Collect the result data. Replace HRDResultColumnId with variable id's
        SQL = """
        SELECT RD.HydraulicLoadId, RV.ResultVariableId, RD.Value
        FROM HydroDynamicResultData RD
        INNER JOIN HRDResultVariables RV ON RD.HRDResultColumnId = RV.HRDResultColumnId
        WHERE HRDLocationId = {};""".format(hrdlocationid)
        resultdata = pd.read_sql(
            SQL, self.con, index_col=["HydraulicLoadId", "ResultVariableId"]
        ).unstack()

        # Reduce columnindex to single level index (without 'Value')
        resultdata.columns = [
            rvids[rid] for rid in resultdata.columns.get_level_values(1)
        ]

        # Create dictionary for mapping HRDInputColumnId to InputVariableId
        SQL = """
        SELECT ID.HydraulicLoadId, IV.InputVariableId, ID.Value
        FROM HydroDynamicInputData ID
        INNER JOIN HRDInputVariables IV ON ID.HRDInputColumnId = IV.HRDInputColumnId"""
        inputdata = pd.read_sql(
            SQL, self.con, index_col=["HydraulicLoadId", "InputVariableId"]
        ).unstack()

        # Reduce columnindex to single level index (without 'Value')
        inputdata.columns = [
            ivids[ivid] for ivid in inputdata.columns.get_level_values(1)
        ]

        # Join data and sort values
        resultaat = (
            dataids.join(inputdata).join(resultdata).sort_values(by=["r", "u", "m"])
        )

        # In the WBI2023 the water levels and waves are in the same table, but have different input variables
        # Split the water levels and wave results
        idx = pd.isnull(resultaat["hs"])

        waterlevels = resultaat.loc[idx].dropna(how="all", axis=1)
        waveconditions = resultaat.loc[~idx].dropna(how="all", axis=1)

        # Replace m_os by m for the water level and m for h for the wave conditions
        waterlevels.rename(columns={"m_os": "m"}, inplace=True)
        waveconditions.rename(columns={"m": "h"}, inplace=True)

        return waterlevels, waveconditions



import numpy as np
import pandas as pd

from itertools import product

from ..loading import Loading
from ..loading_model.loading_model import LoadingModel
from ....location import Location
from ....profile.profile import Profile


class LoadingWaveOvertopping(Loading):
    """
    This Loading is used to calculate HBNs
    """

    def __init__(self, location: Location, ws_range: np.ndarray):
        """
        Init the Loading object for the Waves Overtopping

        Parameters
        ----------
        settings : Settings
            The Settings object
        """
        # Inherit the from parent
        super().__init__(location.get_settings())

        # Init
        self.ws_loading = location.get_model().get_loading()
        self.ws_range = ws_range
        self.read_loading()

    def read_loading(self) -> None:
        """
        Read the HR and create Loading Models
        """
        # Bepaal bij welke waterstanden de golfcondities afgeleid moeten worden
        # Wanneer het aantal unieke waterstanden kleiner is dan het aantal waterstanden
        # in het opgegeven bereik, worden de unieke waterstanden gebruikt
        ws_per_r = self.bepaal_kh_waterstanden(self.ws_loading, self.ws_range)

        # Bepaal de golfcondities bij deze waterstanden
        for richting, waterstanden in ws_per_r.items():
            # Leidt golfcondities af voor een bepaalde richting
            golfcond_r, windsnelheden = self.ws_loading.get_wave_conditions(
                richting, waterstanden, extrapolate=True
            )

            # Create a DataFrame
            df = pd.DataFrame(columns=["wlev", "u"] + list(golfcond_r.keys()))

            # Iterate over the windspeed and waterlevel arrays
            ws_u = np.array(list(product(waterstanden, windsnelheden)))
            golf = np.array(
                [golfcond_r[key].ravel() for key in list(golfcond_r.keys())]
            ).T
            df = pd.DataFrame(
                np.concatenate((ws_u, golf), axis=1),
                columns=["wlev", "u"] + list(golfcond_r.keys()),
            )

            # Create a Loading Model
            model = LoadingModel(richting, None, ["wlev", "u"], list(golfcond_r.keys()))
            model.initialise(df)

            # Save the model
            self.model[(richting, model.closing_situation)] = model

        # Extend the loading models
        self._extend_loadingmodels()

    def calculate_hbn(
        self,
        profile: Profile,
        qcrit: float = 10,
        factor_hs: float = 1.0,
        factor_tspec: float = 1.0,
    ) -> None:
        """
        Add hbn result variables to each of the LoadingModels.
        If 'hbn' is already defined, it will overwrite the old result variable.

        Parameters
        ----------
        profile : Profile
            The profile
        qcrit : float
            The critical discharge
        factor_hs : float
            Factor for the significant wave height, used for model uncertainty
        factor_tspec : float
            Factor for the spectral wave period, used for model uncertainty
        """
        for _, model in self.iter_models():
            model.calculate_hbn(profile, qcrit, factor_hs, factor_tspec)

    def bepaal_kh_waterstanden(self, ws_belasting, ws_range):
        """
        Bepaal waterstanden waarvoor kruinhoogtes berekend moeten worden.

        Per richting wordt bepaald of het aantal unieke waterstanden
        kleiner is dan het aantal waterstanden in een gediscretiseerde range.

        Als dit het geval is, kies de unieke waterstanden. Zo niet, gebruik de range.
        """
        ws_per_r = {r: [] for r in ws_belasting.r}
        for (richting, _), model in ws_belasting.iter_models():
            ws_per_r[richting] += np.unique(model.h).tolist()

        for richting, waterstanden in ws_per_r.items():
            h_all = np.array(waterstanden)[
                np.unique(np.round(waterstanden, 3), return_index=True)[1]
            ]
            # Als er minder waterstanden voorkomen bij deze richting, dan de voorgenomen
            # range, kies dan deze waterstanden
            if len(h_all) <= len(ws_range):
                ws_per_r[richting] = h_all
            else:
                ws_per_r[richting] = ws_range

        return ws_per_r



import numpy as np

from scipy.stats import norm

from ....settings.settings import Settings
from .....common.enum import WaveShapeType
from .....common.interpolate import Interpolate, InterpStruct
from .....common.probability import ProbabilityFunctions
from .....io.file_hydranl import FileHydraNL


class WaveShape:
    """
    Class to describe the wave form statistics.
    Used in describing wave forms in river discharge and lake level.
    """

    def __init__(self, settings: Settings, type: WaveShapeType):
        """
        Constructor class for the WaveShape statistics.

        Parameters
        ----------
        settings : Settings
            The Settings object
        """
        # Placeholder
        self.wave_shapes = None

        # Obtain the parameters
        self.base_duration = settings.base_duration
        self.pw = settings.waveshape_pw
        if type == WaveShapeType.DISCHARGE:
            self.top_duration = settings.top_duration_q
            self.lower_limit = settings.q_min
            self.upper_limit = settings.q_limit
            self.ifh = settings.ifh_q
            self.ifb = settings.ifb_q
        elif type == WaveShapeType.LAKE_LEVEL:
            self.top_duration = settings.top_duration_a
            self.lower_limit = settings.a_min
            self.upper_limit = settings.a_limit
            self.ifh = settings.ifh_a
            self.ifb = settings.ifb_a
        else:
            raise TypeError("[ERROR] Unknown WaveShape type.")

        # Time
        self.time = np.arange(
            0.0, settings.base_duration + 1e-6, settings.waveshape_time_step
        )
        self.ntime = len(self.time)
        if settings.base_duration != max(self.time):
            settings.base_duration = max(self.time)
            print(
                f"[NOTE] Base duration adjusted to blok wind ({settings.waveshape_time_step}h)."
            )

    def initialise_wave_shapes(self, peak: list, climate_change: float = 0.0):
        """
        Initialize the wave shapes for discharge or lake level.

        Parameters
        ----------
        peak : list
            A list with peak values or the discharge / lake level
        climate_change : float
            Increase of discharge or lake level to account for climate change (default = 0.0)
        """
        # Save
        self.peak = peak
        self.npeak = len(peak)

        # Determine shape of the wave
        if not self.pw:
            # Init wave shapes
            self.wave_shapes = self.__initialise_wave_shape(climate_change)
            self.wave_shapes = np.maximum(self.wave_shapes, self.lower_limit)

        # Read from a file
        else:
            raise NotImplementedError()

    def transition_wave(self, transition: float) -> None:
        """
        Transition wave form

        Parameters
        ----------
        transition : float
            Transition in hours
        """
        tijdenlang = np.zeros(2 * self.ntime - 1)
        golflang = np.zeros(2 * self.ntime - 1)

        if transition >= 0.0:
            # Als de verschuiving positief is in de tijd, dan wordt de tijdas
            # vooruit verschoven en eenzelfde periode wordt voor deze
            # verschoven tijdas geplaatst
            it = np.arange(self.ntime)
            tijdenlang[it + self.ntime - 1] = self.time[it] + transition

            it = np.arange(self.ntime - 1)
            tijdenlang[it] = tijdenlang[it + self.ntime - 1] - self.base_duration

        else:
            # Als de verschuiving negatief is in de tijd, dan wordt de tijdas
            # terug verschoven en eenzelfde periode wordt achter deze verschoven
            # tijdas geplaatst
            it = np.arange(self.ntime)
            tijdenlang[it] = self.time[it] + transition
            it = np.arange(1, self.ntime)
            tijdenlang[it + self.ntime - 1] = tijdenlang[it] + self.base_duration

        #  Bij de verschoven en uitgebreide tijdas worden de golfvormen twee keer gezet
        #  Door interpolatie worden op het reguliere rooster de uiteindelijke
        #  verschoven golfvormen verkregen
        intstr = InterpStruct(x=self.time, xp=tijdenlang)
        for ip in range(self.npeak):
            golflang[: self.ntime] = self.wave_shapes[:, ip]
            golflang[self.ntime :] = golflang[1 : self.ntime]
            self.wave_shapes[:, ip] = intstr.interp(fp=golflang, extrapolate=False)

    def get_wave_shapes(self) -> np.ndarray:
        """
        Return the wave shapes.

        Returns
        -------
        np.ndarray
            Array with size (t, peak)
        """
        return self.wave_shapes

    def exceedance_timestamps(self, levels: np.ndarray):
        """
        Determine the exceedance durations for several levels in the waveshapes,
        given the peak levels of the wave.

        Parameters
        ----------
        levels : np.ndarray
            Levels for which the exceedance duration is determined.

        Returns
        -------
        np.ndarray
            Exceedance duration [N peak x N levels]
        """
        # Check if the wave shapes are defined
        if self.wave_shapes is None:
            raise ValueError("[ERROR] Wave shaped are not defined.")

        # Init empty array
        ovduur = np.zeros((self.npeak, len(levels)))

        # Bepaal voor elk blokniveau de overschrijdingsduren
        for ip in range(self.npeak):
            # Get one wave shape
            wave_shape = self.wave_shapes[:, ip]

            # Initieer lege arrays
            t_boven = np.zeros_like(levels)
            t_onder = np.zeros_like(levels)

            # Bepaal waar de niveaus gelijk of hoger zijn aan de golfvorm, hiervoor zijn er geen overschrijdingen
            idx = levels > wave_shape.max()
            t_boven[idx] = -1.0
            t_onder[idx] = -1.0

            # Aangenomen wordt dat de golf een monotoon opgaande en neergaande flank heeft.
            # Onder deze aanname kan met interpolatie de kruisingen bepaald worden.
            diff = wave_shape[1:] - wave_shape[:-1]

            # Het verschil is de afstand van het volgende punt ten opzichte van het huidige punt
            opgaand = np.concatenate([[True], diff > 0])
            neergaand = np.concatenate([diff < 0, [True]])
            t_boven[~idx] = np.interp(
                x=levels[~idx], xp=wave_shape[opgaand], fp=self.time[opgaand]
            )
            t_onder[~idx] = np.interp(
                x=levels[~idx],
                xp=wave_shape[neergaand][::-1],
                fp=self.time[neergaand][::-1],
            )
            ovduur[ip, :] = t_onder - t_boven

        # Return
        return ovduur

    def instantaneous_exceedance_probability(self, exceedance_peaks, levels):
        """
        Calculate the instantaneous exceedance probability of the discharge/water level
        for each peak value of a discharge/water level wave.
        {
            Calculate the time when the discharge/water level is higher than a given level
                and determine the time when the discharge/water level is lower than the given level.
            Determine the duration of exceedance.
            Use the composite trapezoidal rule for calculating the instantaneous exceedance probability
                of the discharge/water level.
        }
        """
        # Calculate the vector with probability densities of the discharge/water level
        peak_probability = ProbabilityFunctions.probability_density(
            self.peak, exceedance_peaks
        ).probability
        instantaneous_prob = np.zeros_like(levels)
        exceedance_duration = self.exceedance_timestamps(levels)

        for ib in range(len(levels)):
            integral = (exceedance_duration[:, ib] * peak_probability).sum()

            # Calculate the instantaneous exceedance probability of the discharge/water level
            instantaneous_prob[ib] = integral / self.base_duration

        return instantaneous_prob

    def cum_norm_s_naar_y(self, y, mu, sigma):
        """
        Bepaal de cumulatieve verdelingsfunctie voor het meerpeil onder de transformatie van s naar y = K(s)
        {
            y= Parameterwaarde
            mu= Gemiddelde van de normale verdeling
            sigma= Standaarddeviatie van de normale verdeling
        }
        """
        if sigma <= 0:
            # Geef foutmelding
            raise ValueError("Sigma <= 0")

        # Integratiegrenzen en stapgrootte
        stap = 0.05
        x = np.arange(0, 20 + 0.1 * stap, stap)

        # Voor de parameterwaarde wordt de integraal over het rooster berekend. Let wel: delta = 0
        yn = (y[None, :] - x[:, None] - mu) / sigma

        dx = np.diff(np.r_[x[0], (x[1:] + x[:-1]) / 2.0, x[-1]])
        f_y_sigma = (np.exp(-x[:, None]) * norm().cdf(x=yn) * dx[:, None]).sum(0)

        return f_y_sigma

    def __initialise_wave_shape(self, climate_change: float) -> np.ndarray:
        """
        Create the wave shapes for discharge or lake level.

        Parameters
        ----------
        climate_change : float
            Increase of discharge or lake level to account for climate change (default = 0.0)

        Returns
        -------
        np.ndarray
            A 2D array with wave shapes (time, q_m_peak)
        """
        # Read the table with top durations (t_top) given q or m (q_m)
        _q_m, _t_top = FileHydraNL.read_file_2columns(self.top_duration)

        # Correct for climate change (increase discharge or lake level)
        if climate_change > 0.0:
            _q_m += climate_change

        # Check if the top duration is larger than the base duration
        if (_t_top > self.base_duration).any():
            print(
                f"[WARNING] Top duration from file '{self.top_duration}' is larger than base duration."
            )

        # Increase the arrays with 1 element to allow for extrapolation
        q_m = np.r_[_q_m, 2.0 * _q_m[-1] - _q_m[-2]]
        t_top = np.r_[_t_top, _t_top[-1]]

        # Calculate for each peak the top duration
        top_durations = Interpolate.inextrp1d(x=self.peak, xp=q_m, fp=t_top)

        # Create an empty array of size (t, peak) for the wave shapes
        wave_shapes = np.zeros((len(self.time), len(self.peak)))

        # Create a trapezoid wave shape
        for ip, top_duration in enumerate(top_durations):
            # If the discretisation step for wave shapes is equal to the base_duration
            if (
                self.peak[ip] < q_m[0]
                or (self.time[1] - self.time[0]) == self.base_duration
            ):
                # Vul de vector met tijdstippen
                ttabel = [0.0, self.base_duration]

                # Vul de vector met golfwaarden
                gtabel = [self.peak[ip], self.peak[ip]]

            # If the top duration is equal to 0.0
            elif top_duration == 0.0:
                # Als ingesnoerd wordt totaan het tijdstip waarop de golfvorm de maximale waarde heeft
                if self.ifb == 0.0:
                    #   Vul de vector met tijdstippen
                    ttabel = np.array(
                        [
                            0,
                            self.base_duration / 2.0 - 0.0001,
                            self.base_duration / 2.0,
                            self.base_duration / 2.0 + 0.0001,
                            self.base_duration,
                        ]
                    )

                    #   Vul de vector met golfwaarden
                    gtabel = np.array(
                        [
                            self.lower_limit,
                            self.lower_limit
                            + self.ifh * (self.peak[ip] - self.lower_limit),
                            self.peak[ip],
                            self.lower_limit
                            + self.ifh * (self.peak[ip] - self.lower_limit),
                            self.lower_limit,
                        ]
                    )

                # Als de golfvorm niet ingesnoerd wordt
                elif self.ifh == 1.0:
                    # Vul de vector met tijdstippen
                    ttabel = np.array(
                        [0.0, self.base_duration / 2.0, self.base_duration]
                    )

                    # Vul de vector met golfwaarden
                    gtabel = np.array(
                        [self.lower_limit, self.peak[ip], self.lower_limit]
                    )

                # Als er sprake is van een blokgolf
                elif self.ifb == (1.0 / (1.0 - self.ifh)):
                    # Vul de vector met tijdstippen
                    ttabel = np.array(
                        [0.0, self.base_duration / 2.0, self.base_duration]
                    )

                    # Vul de vector met golfwaarden
                    gtabel = np.array(
                        [
                            self.lower_limit
                            + self.ifh * (self.peak[ip] - self.lower_limit),
                            self.peak[ip],
                            self.lower_limit
                            + self.ifh * (self.peak[ip] - self.lower_limit),
                        ]
                    )

                # Als de golfvorm wel ingesnoerd wordt
                else:
                    ttabel = np.zeros(5)
                    gtabel = np.zeros(5)

                    # Vul de vector met tijdstippen
                    ttabel[0] = 0.0
                    ttabel[2] = self.base_duration / 2.0
                    ttabel[4] = self.base_duration
                    ttabel[1] = ttabel[0] + self.ifh * (ttabel[2] - ttabel[0])
                    ttabel[1] = ttabel[2] - self.ifb * (ttabel[2] - ttabel[1])
                    ttabel[3] = ttabel[4] + self.ifh * (ttabel[2] - ttabel[4])
                    ttabel[3] = ttabel[2] + self.ifb * (ttabel[3] - ttabel[2])

                    # Vul de vector met golfwaarden
                    gtabel = np.array(
                        [
                            self.lower_limit,
                            self.lower_limit
                            + self.ifh * (self.peak[ip] - self.lower_limit),
                            self.peak[ip],
                            self.lower_limit
                            + self.ifh * (self.peak[ip] - self.lower_limit),
                            self.lower_limit,
                        ]
                    )

            # Als de invoergeg.topduur ongelijk aan nul is
            else:
                # Als de invoergeg.topduur gelijk is aan de invoergeg.basisduur
                if top_duration == self.base_duration:
                    # Vul de vector met tijdstippen
                    ttabel = np.array([0.0, self.base_duration])

                    # Vul de vector met golfwaarden
                    gtabel = np.array([self.peak[ip], self.peak[ip]])

                # Als ingesnoerd wordt totaan het tijdstip waarop de golfvorm de maximale waarde heeft
                elif self.ifb == 0.0:
                    # n = 6

                    # Vul de vector met tijdstippen
                    ttabel = np.array(
                        [
                            0.0,
                            self.base_duration / 2.0 - top_duration / 2.0 - 0.0001,
                            self.base_duration / 2.0 - top_duration / 2.0,
                            self.base_duration / 2.0 + top_duration / 2.0,
                            self.base_duration / 2.0 + top_duration / 2.0 + 0.0001,
                            self.base_duration,
                        ]
                    )

                    # Vul de vector met golfwaarden
                    gtabel = np.array(
                        [
                            self.lower_limit,
                            self.lower_limit
                            + self.ifh * (self.peak[ip] - self.lower_limit),
                            self.peak[ip],
                            self.peak[ip],
                            self.lower_limit
                            + self.ifh * (self.peak[ip] - self.lower_limit),
                            self.lower_limit,
                        ]
                    )

                # Als de golfvorm niet wordt ingesnoerd
                elif self.ifh == 1.0:
                    # Vul de vector met tijdstippen
                    ttabel = np.array(
                        [
                            0.0,
                            self.base_duration / 2.0 - top_duration / 2.0,
                            self.base_duration / 2.0 + top_duration / 2.0,
                            self.base_duration,
                        ]
                    )

                    # Vul de vector met golfwaarden
                    gtabel = np.array(
                        [
                            self.lower_limit,
                            self.peak[ip],
                            self.peak[ip],
                            self.lower_limit,
                        ]
                    )

                # Als er sprake is van een blokgolf
                elif self.ifb == (1.0 / (1.0 - self.ifh)):
                    # Vul de vector met tijdstippen
                    ttabel = np.array(
                        [
                            0.0,
                            self.base_duration / 2.0 - top_duration / 2.0,
                            self.base_duration / 2.0 + top_duration / 2.0,
                            self.base_duration,
                        ]
                    )

                    # Vul de vector met golfwaarden
                    gtabel = np.array(
                        [
                            self.lower_limit
                            + self.ifh * (self.peak[ip] - self.lower_limit),
                            self.peak[ip],
                            self.peak[ip],
                            self.lower_limit
                            + self.ifh * (self.peak[ip] - self.lower_limit),
                        ]
                    )

                # Als de golfvorm wel ingesnoerd wordt
                else:
                    ttabel = np.zeros(6)
                    gtabel = np.zeros(6)

                    # Vul de vector met tijdstippen
                    ttabel[0] = 0.0
                    ttabel[2] = self.base_duration / 2.0 - top_duration / 2.0
                    ttabel[3] = self.base_duration / 2.0 + top_duration / 2.0
                    ttabel[5] = self.base_duration
                    ttabel[1] = ttabel[0] + self.ifh * (ttabel[2] - ttabel[0])
                    ttabel[1] = ttabel[2] - self.ifb * (ttabel[2] - ttabel[1])
                    ttabel[4] = ttabel[5] + self.ifh * (ttabel[3] - ttabel[5])
                    ttabel[4] = ttabel[3] + self.ifb * (ttabel[4] - ttabel[3])

                    # Vul de vector met golfwaarden
                    gtabel = np.array(
                        [
                            self.lower_limit,
                            self.lower_limit
                            + self.ifh * (self.peak[ip] - self.lower_limit),
                            self.peak[ip],
                            self.peak[ip],
                            self.lower_limit
                            + self.ifh * (self.peak[ip] - self.lower_limit),
                            self.lower_limit,
                        ]
                    )

            # Interpoleer de golfvorm naar de gewenste tijdas
            wave_shapes[:, ip] = np.interp(self.time, ttabel, gtabel)

        # Top de golfvormen af op het gewenste niveau
        wave_shapes = np.minimum(wave_shapes, self.upper_limit)

        # Return wave shapes
        return wave_shapes

    @staticmethod
    def bepaal_gezamenlijke_overschrijding(
        golfvormen_st1, niveaus_st1, golfvormen_st2, niveaus_st2
    ) -> np.ndarray:
        # Bepaal overschrijdingstijdstippen van beide stochasten in de golfvorm
        kruisng = np.zeros((len(niveaus_st1), len(niveaus_st2), 4))
        helling = np.zeros((len(niveaus_st1), len(niveaus_st2), 4), dtype=int)
        duren = np.zeros(
            (
                golfvormen_st1.npeak,
                golfvormen_st2.npeak,
                len(niveaus_st1),
                len(niveaus_st2),
            ),
            dtype=float,
        )

        # Bepaal eerst alle tijdstippen voor stochast 1
        kruisingen_st1 = np.zeros(
            (golfvormen_st1.npeak, len(niveaus_st1), 2), dtype=float
        )
        helling_st1 = np.zeros((golfvormen_st1.npeak, len(niveaus_st1), 2), dtype=int)
        for ip1 in range(golfvormen_st1.npeak):
            # Bepaal voor elk blokniveau de overschrijdingsduren
            kruisingen_st1[ip1, :, :], helling_st1[ip1, :, :] = (
                golfvormen_st1._overschrijdingstijdstip_op_af_v2(
                    golfvormen_st1.wave_shapes[:, ip1], golfvormen_st1.time, niveaus_st1
                )
            )

        arange = np.arange(max(len(niveaus_st1), len(niveaus_st2)))

        # loop vervolgens over stochast 2
        kruisingen_st2 = np.zeros((len(niveaus_st2), 2))
        helling_st2 = np.zeros((len(niveaus_st2), 2))
        for ip2 in range(golfvormen_st2.npeak):
            # Bepaal voor elk blokniveau de overschrijdingsduren
            kruisingen_st2[:, :], helling_st2[:, :] = (
                golfvormen_st2._overschrijdingstijdstip_op_af_v2(
                    golfvormen_st2.wave_shapes[:, ip2], golfvormen_st2.time, niveaus_st2
                )
            )

            # Bepaal voor elk piekniveau van stochast 2 de gezamenlijke overschrijdingen
            # Doe dit door de tijdstippen voor elke combinatie samen te voegen, te sorteren
            # op absolute waarde, en de overgang van laatste opgaand naar eerste neergaand
            # te bepalen
            for ip1 in range(golfvormen_st1.npeak):
                idx1 = kruisingen_st1[ip1, :, 0] != -1.0
                idx2 = kruisingen_st2[:, 0] != -1.0

                len1 = idx1.sum()
                len2 = idx2.sum()
                kruisng = np.zeros((len1, len2, 4))
                helling = np.zeros((len1, len2, 4), dtype=np.int8)

                # Vul arrays
                kruisng[:, :, :2] = kruisingen_st1[ip1, idx1, None, :]
                kruisng[:, :, 2:] = kruisingen_st2[None, idx2, :]

                helling[:, :, :2] = helling_st1[ip1, idx1, None, :]
                helling[:, :, 2:] = helling_st2[None, idx2, :]

                # Sorteer
                order = np.argsort(np.absolute(kruisng), axis=2)
                i1, i2 = arange[:len1, None, None], arange[None, :len2, None]
                kruisng[...] = kruisng[i1, i2, order]
                helling[...] = helling[i1, i2, order]

                # Bepaal alle plekken waar een tekenwisseling (kruising tussen golf en niveau) plaatsvindt
                wh = np.where((helling[:, :, 1:] ^ helling[:, :, :-1]))

                # Bereken het tijdsverschil bij tekenwisseling
                diff = kruisng[wh[0], wh[1], wh[2] + 1] - kruisng[wh[0], wh[1], wh[2]]
                # Zet tijdverschil bij neergaand naar opgaand op 0
                diff[
                    (helling[wh[0], wh[1], wh[2] + 1] < helling[wh[0], wh[1], wh[2]])
                ] = 0.0
                # print(diff.max(), diff.min())
                duren[ip1, ip2, wh[0], wh[1]] += diff

        return duren

    @staticmethod
    def _overschrijdingstijdstip_op_af_v2(
        vormgolf: np.ndarray, tijden: np.ndarray, niveaus: np.ndarray
    ):
        # Bepaal waar de golfvorm boven het te testen niveau zit
        diff = (vormgolf[None, :] - niveaus[:, None]) >= 0.0
        # Bepaal alle plekken waar een tekenwisseling (kruising tussen golf en niveau) plaatsvindt
        locn, loct = np.where((diff[:, 1:] ^ diff[:, :-1]))

        # Bepaal het snijpunt bij de kruising
        dtdq = (tijden[loct + 1] - tijden[loct]) / (vormgolf[loct + 1] - vormgolf[loct])
        # Voeg dit punt toe aan de tijd
        duur = tijden[loct] + dtdq * (niveaus[locn] - vormgolf[loct])

        # Maak lege array aan
        t1t2 = np.full((len(niveaus), 2), np.nan, dtype=float)
        helling = np.full((len(niveaus), 2), 0, dtype=np.int8)
        # Waar het niveau hoger ligt dan de piekwaarde, -1 in
        t1t2[niveaus > vormgolf.max(), :] = -1
        # Controleer of alle niveaus of 0 of 2 keer voorkomen
        r = np.bincount(locn)
        assert ((r == 2) | (r == 0)).all()

        # Vul in in array met tijden. Vermenigvuldig met het teken van de helling *-1
        t1t2[locn, np.arange(len(locn)) % 2] = duur
        # Bepaal helling. Opgaand heeft dus -, neergaand +. De som geeft de totale overschrijdingsduur
        helling[locn, np.arange(len(locn)) % 2] = -np.sign(dtdq)

        # Als er NaN's in de array zitten, is dit op plekken waar het niveau in zijn geheel onder
        # of op het laagste punt van de golf ligt. Vul hier de min en max van de tijden in
        idx = np.isnan(t1t2)
        t1t2[idx[:, 0], 0] = tijden.min()
        t1t2[idx[:, 1], 1] = tijden.max()
        helling[idx[:, 0], 0] = -1
        helling[idx[:, 1], 1] = 1

        # Vermenigvuldig de even elementen met -1
        # idx = duur < 0
        # duur[::2] *= -1
        return t1t2, helling
